/*
    This file is part of LWX. 

    LWX is the abbreviation of the Light WiMax simulation program which 
    is a ns2 WiMAX network simulation module. LWX allows each user to 
    apply his/her designed scheduling algorithm or other mechanism used 
    in WiMAX network and generate the corresponding simulation results
    for further analysis or other non-commercial purposes.

    Copyright (C) 2008 Yen-Hung Cheng (pplong), NTUST in Taiwan, R.O.C.
    E-Mail: lwx.ns2@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "lwx_mac.h"
#include "connector.h"
#include "lwx_bind_tcl.cc"
#include "lwx_tool.cc"
#include <sstream>



///////////////////
// init
///////////////////

// init lwx mac layer
LWX::LWX() : Mac() 
{
  // adding a new timer
  frame_timer = new Frame_LWXTimer(this);
  rtg_timer = new RTG_LWXTimer(this);
  
  // timer start while the simulation is start
  frame_timer->start(0);
  
}

int LWX::command(int argc, const char*const* argv)
{
  
	if (argc == 3) 
	{
		if(strcmp(argv[1], "eventtrace") == 0) 
		{
			return (TCL_OK);
		}

	}
	return Mac::command(argc, argv);
}


// init lwx pkt header

int hdr_lwx::offset_;


///////////////////
// receiving packet from the upper layer or lower layer
///////////////////

void LWX::recv(Packet *p, Handler *h) {

  hdr_cmn *cmn_col  = hdr_cmn::access(p);
  hdr_mac *mac_col  = hdr_mac::access(p);
  hdr_lwx *lwx_col  = hdr_lwx::access(p);
  hdr_ip  *ip_col   = hdr_ip::access(p);  
  
  
  // 20080623 test rs start
  // lwx header log
  /*
  hdr_lwx  *lwx_col  = hdr_lwx::access(p);
  if(cmn_col->ptype()==PT_LWX)
  {  

    ostringstream test_rs_msg;
	  p_info *test_rs_pkt_info;
	  test_rs_msg << "get ctrl pkt from mac: 0"
	              << " packet type " <<  test_rs_pkt_info->name(cmn_col->ptype())
                << " src_node_id " << Node::get_node_by_address(index_)->nodeid()
                << " recv_node_id " << index_
                << " mac_src " << mac_col->macSA_ 
                << " mac_dst " << mac_col->macDA_
                << " src_ip " << ip_col->saddr()
                << " src_port " << ip_col->sport()
                << " dst_ip " << ip_col->daddr()
                << " dst_port " << ip_col->dport() 
                << " lwx_a " << lwx_col->a()
                << " lwx_b " << lwx_col->b()
                << " lwx_c " << lwx_col->c()
                << " lwx_d " << lwx_col->d()
                << endl;
    lwx_test_msg ("recv", test_rs_msg.str());
    return;
  }
  */


  // send down packet start
  if (cmn_col->direction() == hdr_cmn::DOWN) 
  {
    // o@OI, ڤDn[oF~|d, ڷQ, ӬO~MAC]
    callback_ = h;
    
    // 20080624 test rs start
    /*
    sendDown(p);
    return;
    */
    // 20080624 test rs end
    
      
    //  ]~MACAFeLMAC messageAno@qAMʥ]|d...
    //  ]ApGJ즹ʥ]AXhAM}禡
    //
    //   MI, Ъ`N:
    //     ثeOeARPAODV, pGo͵{d, 
    //     iOLmac packetSQofunction
    //     Х[HT{

    // for log start
    ostringstream tmp_down_msg;
	  p_info *test_rs_pkt_info;
	  tmp_down_msg  << " packet type " <<  test_rs_pkt_info->name(cmn_col->ptype())
                << " recv_node_id " << Node::get_node_by_address(index_)->nodeid()
                << " recv_mac " << index_
                << " mac_src " << mac_col->macSA_ 
                << " mac_dst " << mac_col->macDA_
                << " src_ip " << ip_col->saddr()
                << " src_port " << ip_col->sport()
                << " dst_ip " << ip_col->daddr()
                << " dst_port " << ip_col->dport() 
                << " lwx_bwa_flow_type " << lwx_col->bwa_flow_type();
    //lwx_test_msg("recv_down", tmp_down_msg.str());
    // for log end

    if(cmn_col->ptype()==PT_AODV ||
       cmn_col->ptype()==PT_ARP)
    {

      sendDown(p);
      return;
    }
       
    //
    // Dmac layerʥ]i۹flow packet queue
    //
 
    // find the position of the packet flow in the lwx_flow_list
    int src_node_id   = Node::get_node_by_address(index_)->nodeid();
    int src_ip        = ip_col->saddr();
    int src_port      = ip_col->sport();
    int dst_ip        = ip_col->daddr();
    int dst_port      = ip_col->dport();
    
    // find its belonging node id
    for(int i=0 ;i <lwx_node_list.size();i++)
    {
      if(lwx_node_list.at(i).node_id == src_node_id )
      {
        for(int j=0;j < lwx_node_list.at(i).ctrl_flow.size();j++)
        {
          Flow_Queue *tmp_flow_queue = &lwx_node_list.at(i).ctrl_flow.at(j);
          
          if(tmp_flow_queue->src_node_id == src_node_id &&
               tmp_flow_queue->src_ip == src_ip &&
               tmp_flow_queue->src_port == src_port &&
               tmp_flow_queue->dst_ip == dst_ip &&
               tmp_flow_queue->dst_port == dst_port)
          {
            // set lwx packet header column
            
            lwx_col->src_ip()          = ip_col->saddr();
            lwx_col->src_port()        = ip_col->sport();
            lwx_col->src_mac()         = mac_col->macSA_;
            lwx_col->dst_ip()          = ip_col->daddr();
            lwx_col->dst_port()        = ip_col->dport();
            lwx_col->dst_mac()         = mac_col->macDA_;
            lwx_col->src_node_id()     = tmp_flow_queue->src_node_id;
            lwx_col->from_node_id()    = tmp_flow_queue->from_node_id;
            lwx_col->to_node_id()      = tmp_flow_queue->to_node_id;
            lwx_col->bwa_flow_type()   = tmp_flow_queue->flow_type;
            lwx_col->sendtime()        = Scheduler::instance().clock();
            lwx_col->maxl()            = tmp_flow_queue->Lmax; //pandeli

            //packetsiqueue
            tmp_flow_queue->pkt_push(p);
            
            // sUMAChBzoӫʥ]ƥ(ëDsUMACʥ]ǥXh)
            Scheduler& s = Scheduler::instance();				
            s.schedule(&hRes_, &intr_, 0);
           
            return;
          }
        }
      }
	  }
	  
	  // pG]o̡ANpacketSflow
	 

	  string tmp_where   = "LWX::recv() in if (cmn_col->direction() == hdr_cmn::DOWN)";
    lwx_err_msg(tmp_where);
    
  }
  // send down packet end

  // send up packet start
  if (cmn_col->direction() == hdr_cmn::UP) 
  {
      
    //  ]~MACAFeLMAC messageAno@qAMʥ]|d...
    //  ]ApG즹ʥ]AWAM}禡
    //
    //   MI, Ъ`N:
    //     ثeOeARPAODV, pGoͨSʥ]iӡAάO{`, 
    //     iOLmac packetSQofunction
    //     Х[HT{
    if(cmn_col->ptype()==PT_AODV ||
       cmn_col->ptype()==PT_ARP)
    {

      sendUp(p);
      return;
    }
    

    // drop the packet whose mac->macDA_ is not this node's mac addr
    if(mac_col->macDA_ != index_)
    {
      Packet::free(p);
      return;
    }
///////////////////////////////////////////////////////////////////////////////
            
  /* // drop the packet if it is expired (expirity is defined by Lmax) -Pandeli
   if (Scheduler::instance().clock() - lwx_col->sendtime_ > lwx_col->maxl_ )
    { 
     cout << "src_IP: " << ip_col->saddr() << " src_port:" << ip_col->sport()
          << "dest_IP:" << ip_col->daddr() << "dest_port:"<< ip_col->dport() 
          << "Lmax:" << lwx_col->maxl_ << "sendtime:" << lwx_col->sendtime_  
          << endl ; // prints Output sentence on screen    
     drop(p);
     Packet::free(p);     		 
     return;
     }*/

////////////////////////////////////////////////////////////////////////////////////////////////

    // if this packet's bwa_flow_type is bs->ss or ss->bs, 
    //                                   rs_to_bs, rs_to_ss_tp, rs_to_ss_nontp
    // sendup directly
    else if( (lwx_col->bwa_flow_type() == FLOW_TYPE_BS_TO_SS) ||
             (lwx_col->bwa_flow_type() == FLOW_TYPE_SS_TO_BS) ||
             (lwx_col->bwa_flow_type() == FLOW_TYPE_RS_TO_BS) ||
             (lwx_col->bwa_flow_type() == FLOW_TYPE_RS_TO_SS_TP) ||
             (lwx_col->bwa_flow_type() == FLOW_TYPE_RS_TO_SS_NONTP)) //pandeli
    {
      sendUp(p);
      return;
    }
    else if((lwx_col->bwa_flow_type() == FLOW_TYPE_BS_TO_RS) ||
            (lwx_col->bwa_flow_type() == FLOW_TYPE_SS_TO_RS_TP) ||
            (lwx_col->bwa_flow_type() == FLOW_TYPE_SS_TO_RS_NONTP) ||
             (lwx_col->bwa_flow_type() == FLOW_TYPE_RS_TO_RS_NONTP_UL) ||
             (lwx_col->bwa_flow_type() == FLOW_TYPE_RS_TO_RS_NONTP_DL)||
             (lwx_col->bwa_flow_type() == FLOW_TYPE_RS_TO_RS_NONTP_UL_ODD)||
             (lwx_col->bwa_flow_type() == FLOW_TYPE_RS_TO_RS_NONTP_DL_ODD)) //pandeli
    {
      // find this pkt arriving node's node id
      int pkt_arriving_rs_nid   = Node::get_node_by_address(index_)->nodeid();
        
      // get this pkt's flow info
      int pkt_src_ip        = lwx_col->src_ip();
      int pkt_src_port      = lwx_col->src_port();
      int pkt_dst_ip        = lwx_col->dst_ip();
      int pkt_dst_port      = lwx_col->dst_port();
      
      // find its belonging node id
      for(int i=0 ;i <lwx_node_list.size();i++)
      {
        if(lwx_node_list.at(i).node_id == pkt_arriving_rs_nid )
        {
          for(int j=0;j < lwx_node_list.at(i).ctrl_flow.size();j++)
          {
            Flow_Queue *tmp_flow_queue = &lwx_node_list.at(i).ctrl_flow.at(j);
            
            if(  tmp_flow_queue->src_ip == pkt_src_ip &&
                 tmp_flow_queue->src_port == pkt_src_port &&
                 tmp_flow_queue->dst_ip == pkt_dst_ip &&
                 tmp_flow_queue->dst_port == pkt_dst_port)
            {
                               
              	//packetsiqueue
              	tmp_flow_queue->pkt_push(p);
                return;
		
            }
          }
        }
  	  }
  	  
  	  // pG]o̡ANpacketSflow
  	 
    
  	  string tmp_where   = "LWX::recv() in if (cmn_col->direction() == hdr_cmn::UP)";
      lwx_err_msg(tmp_where, "this pkt doesn't find its relay flow queue");
      
    }
        
    lwx_err_msg( "LWX::recv() UP's final section","program shouldn't go through here!");
    
  }
  // send down packet end
  
  
  lwx_err_msg( "LWX::recv() final section");
  
	
}

/////////////////////
//  Bandwidth Allocation
//  => please write down your algorithm in bandwidth_allocation()
/////////////////////

/*
what var you can use for bandwidth allocation

<in lwx_mac.h>

  #define FRAME_PERIOD            0.005
  #define FRAME_SYMBOL            48
  #define DL_PUSC_SUBCH_NUM       60
  #define UL_PUSC_SUBCH_NUM       70
  #define DL_PUSC_SLOT_SYMBOL_NUM 2
  #define UL_PUSC_SLOT_SYMBOL_NUM 2
  
  static vector <LWX_Node> lwx_node_list;
  
what you need to set is
  lwx_node_list.at(i).ctrl.at(j).BWA_Total;
  lwx_node_list.at(i).ctrl.at(j).BWA_FLOW_TYPE;
  lwx_node_list.at(i).ctrl.at(j).BWA_Next_Hop;
*/

// algorithm 0: Single frame RR for nt relay link Pandeli
// algorithm 1: Multi frame RR for nt relay link Pandeli
// algorithm 2: Round Robin (RR)
// algorithm 3: strict priority
// algorithm 101: original relay sample

void LWX::bandwidth_allocation(int node_id)
{
  
  switch(BWA_Algorithm_Type)
  {
    case 0:
      Single_Frame(node_id);
    case 1:
      Multi_Frame(node_id);
      break;
    case 2:
      BWA_Round_Robin(node_id);
      break;
    case 3:
      BWA_Strict_Priority(node_id);  
      break;
    case 101:
      BWA_Relay_Sample(node_id);
      break;
    default:
      BWA_Round_Robin(node_id);
  }

}

void LWX::Single_Frame(int node_id)
{
  /******************************
  * PANDELI'S ALGORITHM single - frame        *
  ******************************/

  // create pointer to each flow type's queue to allocation bandwidth efficiencially
  vector <Flow_Queue*> tmp_dl_access_flow;
  vector <Flow_Queue*> tmp_dl_relay_access_flow;
  vector <Flow_Queue*> tmp_dl_relay_flow;
  vector <Flow_Queue*> tmp_ul_access_flow;
  vector <Flow_Queue*> tmp_ul_relay_access_flow;
  vector <Flow_Queue*> tmp_ul_relay_flow;
  
  
  // get symbol num of DL and UL
  int dl_access_zone_bwa  = DL_ACCESS_ZONE_SYMBOL * DL_PUSC_SUBCH_NUM / AZ; 
  int dl_relay_access_zone_bwa  = DL_ACCESS_ZONE_SYMBOL * DL_PUSC_SUBCH_NUM /AZ * NH; 
  int dl_relay_zone_bwa   = DL_RELAY_ZONE_SYMBOL * DL_PUSC_SUBCH_NUM / Nb_RS; 
  int ul_access_zone_bwa  = UL_ACCESS_ZONE_SYMBOL * UL_PUSC_SUBCH_NUM /AZ;
  int ul_relay_access_zone_bwa  = UL_ACCESS_ZONE_SYMBOL * UL_PUSC_SUBCH_NUM /AZ * NH;
  int ul_relay_zone_bwa   = UL_RELAY_ZONE_SYMBOL * UL_PUSC_SUBCH_NUM / Nb_RS; 
  
  // algorithm calculation var
  int dl_access_zone_bwa_free=0;
  int dl_relay_access_zone_bwa_free=0;
  int dl_relay_zone_bwa_free=0;
  int ul_access_zone_bwa_free=0;
  int ul_relay_access_zone_bwa_free=0; 
  int ul_relay_zone_bwa_free=0;

  
  // reset all relative var's value in flow's attribute
  for(int i=0; i < lwx_node_list.size(); i++)
  {
    LWX_Node *tmp_node = &lwx_node_list.at(i);
    for( int j=0; j <  tmp_node->ctrl_flow.size(); j++)
    { 
      
      Flow_Queue *tmp_flow_queue = &tmp_node->ctrl_flow.at(j);
      tmp_flow_queue->BWA.clear();
      tmp_flow_queue->BWA_Total=-1;
      tmp_flow_queue->BWA_Rate=-1;
      tmp_flow_queue->BWA_Total_Byte=-1;
      tmp_flow_queue->BWA_Flow_Type=-1;
      tmp_flow_queue->BWA_Next_Hop=-1;
    }
  }



  // =========================================================
  // path selection:
  //   1.find the path (directly to bs or passing through rs)
  //   2.set BWA_Flow_Type and BWA_Next_Hop
  // =========================================================
  
  // set relay route's access link (second hop) to its own flow_type
  // Set access zone flow for ss, flow_type is the default flow type
  // e.g. rs->ss_tp 
  for(int i=0; i < lwx_node_list.size(); i++)
  {
    LWX_Node *tmp_node = &lwx_node_list.at(i);
    for( int j=0; j <  tmp_node->ctrl_flow.size(); j++)
    { 
      Flow_Queue *tmp_flow_queue = &tmp_node->ctrl_flow.at(j);
      switch(tmp_flow_queue->flow_type)
      {
        case FLOW_TYPE_RS_TO_SS_TP:
          tmp_flow_queue->BWA_Flow_Type = tmp_flow_queue->flow_type;
          tmp_flow_queue->BWA_Next_Hop  = tmp_flow_queue->to_node_id;
          tmp_flow_queue->BWA_Rate      = tmp_flow_queue->rate;
          break;
        default:
          break;
      }
      
    }
  }
  
  // path selection:
  //   # if this flow has relay link, use the first relay link as its next hop
  //   # else use access link
  for(int i=0; i < lwx_node_list.size(); i++)
  {
    LWX_Node *tmp_node = &lwx_node_list.at(i);
    for( int j=0; j <  tmp_node->ctrl_flow.size(); j++)
    { 

      Flow_Queue *tmp_flow_queue = &tmp_node->ctrl_flow.at(j);
      if(tmp_flow_queue->flow_type == FLOW_TYPE_BS_TO_SS)
      {
        if(!tmp_flow_queue->relay_link.empty())
        {
          tmp_flow_queue->BWA_Flow_Type = tmp_flow_queue->relay_link.at(0).flow_type;
          tmp_flow_queue->BWA_Next_Hop  = tmp_flow_queue->relay_link.at(0).rs_node_id;
          tmp_flow_queue->BWA_Rate      = tmp_flow_queue->relay_link.at(0).rate;

     /* cout << "Temp node: " << tmp_node->node_id << " flow size :" << tmp_node->ctrl_flow.size()
          << "relay link from BS: " << tmp_flow_queue->relay_link.at(0).rs_node_id << endl ; // prints Output sentence on screen */
        }
        else
        {
          tmp_flow_queue->BWA_Flow_Type = tmp_flow_queue->flow_type;
          tmp_flow_queue->BWA_Next_Hop  = tmp_flow_queue->to_node_id;
          tmp_flow_queue->BWA_Rate      = tmp_flow_queue->rate;
        }
      }
      
      if(tmp_flow_queue->flow_type == FLOW_TYPE_SS_TO_BS)
      {
        if(!tmp_flow_queue->relay_link.empty())
        {
          tmp_flow_queue->BWA_Flow_Type = tmp_flow_queue->relay_link.at(0).flow_type;
          tmp_flow_queue->BWA_Next_Hop  = tmp_flow_queue->relay_link.at(0).rs_node_id;
          tmp_flow_queue->BWA_Rate      = tmp_flow_queue->relay_link.at(0).rate;

    /* cout << "Temp node: " << tmp_node->node_id  << " flow size :" << tmp_node->ctrl_flow.size()
          << "relay link from SS: " << tmp_flow_queue->relay_link.at(0).rs_node_id << endl ; // prints Output sentence on screen */
        }
        else
        {
          tmp_flow_queue->BWA_Flow_Type = tmp_flow_queue->flow_type;
          tmp_flow_queue->BWA_Next_Hop  = tmp_flow_queue->to_node_id;
          tmp_flow_queue->BWA_Rate      = tmp_flow_queue->rate;
        }
      }
      
     if((tmp_flow_queue->flow_type == FLOW_TYPE_RS_TO_SS_NONTP)||
        (tmp_flow_queue->flow_type == FLOW_TYPE_RS_TO_BS))
      {
        if(!tmp_flow_queue->relay_link.empty())
        {
          tmp_flow_queue->BWA_Flow_Type = tmp_flow_queue->relay_link.at(0).flow_type;
          tmp_flow_queue->BWA_Next_Hop  = tmp_flow_queue->relay_link.at(0).rs_node_id;
          tmp_flow_queue->BWA_Rate      = tmp_flow_queue->relay_link.at(0).rate;
     
/* cout << "Temp node: " << tmp_node->node_id  << " flow size :" << tmp_node->ctrl_flow.size()
          << "relay link from RS: " << tmp_flow_queue->relay_link.at(0).rs_node_id << endl ; // prints Output sentence on screen */
        }
        else
        {
          tmp_flow_queue->BWA_Flow_Type = tmp_flow_queue->flow_type;
          tmp_flow_queue->BWA_Next_Hop  = tmp_flow_queue->to_node_id;
          tmp_flow_queue->BWA_Rate      = tmp_flow_queue->rate;
          //lwx_err_msg("in relay zone", 
             //"wrong flow type, there is no connection for this relay zone DL BS- to RS");
        }
      }

    }
  }
  
  // =========================================================
  // bandwidth allocation:
  //   # access zone: RR + min QoS guarantee
  //   # relay zone:  only RR
  // =========================================================

  // collect flow into corresponding flow queue
  
  // !!!!!!! very important design: !!!!!!!!!!! Modified to wok for NT RS
  //   bs->ss and rs->ss should be scheduling together, because they use same bandwidth
  //   ss->bs and ss->rs should be also scheduling together, because they use same bandwidth
  
  for(int i=0; i < lwx_node_list.size(); i++)
  {
    LWX_Node *tmp_node = &lwx_node_list.at(i);
    for( int j=0; j <  tmp_node->ctrl_flow.size(); j++)
    { 
      switch(tmp_node->ctrl_flow.at(j).BWA_Flow_Type)
      {
        case FLOW_TYPE_BS_TO_SS:
          //tmp_dl_access_flow.push_back(&tmp_node->ctrl_flow.at(j));
        case FLOW_TYPE_RS_TO_SS_TP:
          tmp_dl_access_flow.push_back(&tmp_node->ctrl_flow.at(j));
            break;
        case FLOW_TYPE_RS_TO_SS_NONTP:
          tmp_dl_relay_access_flow.push_back(&tmp_node->ctrl_flow.at(j));
         // lwx_err_msg("OK BS to RS \n");
            break; //new
        case FLOW_TYPE_BS_TO_RS:
         // tmp_dl_relay_flow.push_back(&tmp_node->ctrl_flow.at(j));
        case FLOW_TYPE_RS_TO_RS_NONTP_DL:
          tmp_dl_relay_flow.push_back(&tmp_node->ctrl_flow.at(j));
           //lwx_err_msg("OK RS to RS \n");
          break; 
        case FLOW_TYPE_SS_TO_RS_TP:
         // tmp_ul_access_flow.push_back(&tmp_node->ctrl_flow.at(j));
        case FLOW_TYPE_SS_TO_BS:
          tmp_ul_access_flow.push_back(&tmp_node->ctrl_flow.at(j));
          break;
        case FLOW_TYPE_SS_TO_RS_NONTP:          
          tmp_ul_relay_access_flow.push_back(&tmp_node->ctrl_flow.at(j));
           break; //new
        case FLOW_TYPE_RS_TO_BS:          
          //tmp_ul_relay_flow.push_back(&tmp_node->ctrl_flow.at(j));
        case FLOW_TYPE_RS_TO_RS_NONTP_UL:
          tmp_ul_relay_flow.push_back(&tmp_node->ctrl_flow.at(j));
         //lwx_err_msg("OK RS to SS \n");
          break;
        default:
          lwx_err_msg("in bandwidth allocation", 
             "wrong flow type, this algorithm doesn't support 802.16j transparent RS");
      }
      
    }
  }
  

  //
  //  downlink access bandwidth allocation
  //


  // satisfy each dl flow's Rmin
  dl_access_zone_bwa_free = dl_access_zone_bwa;
  for( int i=0; i <  tmp_dl_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_dl_access_flow.at(i);
    
    if( tmp_flow_queue->QoS_Class == 1 || tmp_flow_queue->QoS_Class == 2 ||
        tmp_flow_queue->QoS_Class == 3 || tmp_flow_queue->QoS_Class == 4 )
    {
      int allo_symbol = (int)ceil((tmp_flow_queue->Rmin)
                              /(8*PACKETS_PER_SECOND*tmp_flow_queue->BWA_Rate));
      allo_symbol     = (int)ceil ( allo_symbol/ DL_PUSC_SLOT_SYMBOL_NUM) 
                                * DL_PUSC_SLOT_SYMBOL_NUM;
      
      if(allo_symbol > dl_access_zone_bwa_free)
      {
        allo_symbol=dl_access_zone_bwa_free; 
      }
      
      tmp_flow_queue->BWA.push_back(allo_symbol);
      tmp_flow_queue->BWA_Total+=allo_symbol;
      
      dl_access_zone_bwa_free -= allo_symbol;
      
      if(dl_access_zone_bwa_free <0)
      {
        lwx_err_msg("RR's DL Rmin", "dl_free_symbol_num <0");
      }
      
    }
  }
  dl_access_zone_bwa = dl_access_zone_bwa_free;
  
  
  // satisfy each dl flow by RR
  dl_access_zone_bwa_free = dl_access_zone_bwa;
  for( int i=0; i <  tmp_dl_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_dl_access_flow.at(i);
    
    int allo_symbol_max = (int)floor((tmp_flow_queue->Rmax)
                                 /(8*PACKETS_PER_SECOND*tmp_flow_queue->BWA_Rate));
    allo_symbol_max     = (int)ceil ( allo_symbol_max/ DL_PUSC_SLOT_SYMBOL_NUM) 
                                 * DL_PUSC_SLOT_SYMBOL_NUM;
    int allo_symbol = (int)ceil ( dl_access_zone_bwa_free / (tmp_dl_access_flow.size()-i));
    allo_symbol     = (int)ceil ( allo_symbol/ DL_PUSC_SLOT_SYMBOL_NUM) 
                             * DL_PUSC_SLOT_SYMBOL_NUM;
      
    if(allo_symbol_max <= allo_symbol + tmp_flow_queue->BWA_Total)
    {
      allo_symbol = allo_symbol_max - tmp_flow_queue->BWA_Total;
    }
    tmp_flow_queue->BWA_Total += allo_symbol;
      
    dl_access_zone_bwa_free -= allo_symbol;
      
    if(dl_access_zone_bwa_free <0)
    {
      lwx_err_msg("RR's DL RR", "dl_access_zone_bwa_free <0");
    }

  }
  dl_access_zone_bwa = dl_access_zone_bwa_free;
  

  // final dl BWA setting
  for( int i=0; i <  tmp_dl_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_dl_access_flow.at(i);
    
    
    tmp_flow_queue->BWA_Total_Byte = (int)(tmp_flow_queue->BWA_Total 
                                     * tmp_flow_queue->BWA_Rate+ tmp_flow_queue->BWA_Backoff) ;
    tmp_flow_queue->BWA_Backoff = 0 ;
    
  }
  // log each phase's bandwidth allocation of each downlink flow
  ostringstream dl_access_bwa_log;
  dl_access_bwa_log << "============ dl access bwa log: ============" << endl;
  dl_access_bwa_log << "{(src_node_id/from_node_id/to_node_id) (src_ip/src_port/dst_port)\n"
                    << " (flow_type/rate) qsize (queue_size) \n"
                    << "BWA: (Total/Total_Byte/Flow_type/Next_Hop/Rate)} \n";
  
  for( int i=0; i <  tmp_dl_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_dl_access_flow.at(i);
    
    dl_access_bwa_log << "{("
                      << tmp_flow_queue->src_node_id << "/" 
                      << tmp_flow_queue->from_node_id << "/" 
                      << tmp_flow_queue->to_node_id << "/"  
                      << tmp_flow_queue->dst_port << ") " ;
    
    dl_access_bwa_log << "(" << tmp_flow_queue->src_ip << "/" 
                      << tmp_flow_queue->src_port << "/" 
                      << tmp_flow_queue->dst_ip << "/"  
                      << tmp_flow_queue->dst_port << ") ";
                      
    dl_access_bwa_log << "(" << tmp_flow_queue->flow_type << "/" 
                      << tmp_flow_queue->rate << ") ";
    
    dl_access_bwa_log << "qsize (" << tmp_flow_queue->queue_size() << ") " ;
    
    dl_access_bwa_log << "BWA: (" 
                      << tmp_flow_queue->BWA_Total << "/" 
                      << tmp_flow_queue->BWA_Total_Byte << "/" 
                      << tmp_flow_queue->BWA_Flow_Type << "/" 
                      << tmp_flow_queue->BWA_Next_Hop << "/" 
                      << tmp_flow_queue->BWA_Rate << ")} " ;
                      
  }
 //
  //  downlink relay access bandwidth allocation
  //


  // satisfy each dl flow's Rmin
  dl_relay_access_zone_bwa_free = dl_relay_access_zone_bwa;
  for( int i=0; i <  tmp_dl_relay_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_dl_relay_access_flow.at(i);
    
    if( tmp_flow_queue->QoS_Class == 1 || tmp_flow_queue->QoS_Class == 2 ||
        tmp_flow_queue->QoS_Class == 3 || tmp_flow_queue->QoS_Class == 4 )
    {
      int allo_symbol = (int)ceil((tmp_flow_queue->Rmin)
                              /(8*PACKETS_PER_SECOND*tmp_flow_queue->BWA_Rate));
      allo_symbol     = (int)ceil ( allo_symbol/ DL_PUSC_SLOT_SYMBOL_NUM) 
                                * DL_PUSC_SLOT_SYMBOL_NUM;
      
      if(allo_symbol > dl_relay_access_zone_bwa_free)
      {
        allo_symbol=dl_relay_access_zone_bwa_free; 
      }
      
      tmp_flow_queue->BWA.push_back(allo_symbol);
      tmp_flow_queue->BWA_Total+=allo_symbol;
      
      dl_relay_access_zone_bwa_free -= allo_symbol;
      
      if(dl_relay_access_zone_bwa_free <0)
      {
        lwx_err_msg("RR's DL Rmin", "dl_free_symbol_num <0");
      }
      
    }
  }
  dl_relay_access_zone_bwa = dl_relay_access_zone_bwa_free;
  
  
  // satisfy each dl flow by RR
  dl_relay_access_zone_bwa_free = dl_relay_access_zone_bwa;
  for( int i=0; i <  tmp_dl_relay_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_dl_relay_access_flow.at(i);
    
    int allo_symbol_max = (int)floor((tmp_flow_queue->Rmax)
                                 /(8*PACKETS_PER_SECOND*tmp_flow_queue->BWA_Rate));
    allo_symbol_max     = (int)ceil ( allo_symbol_max/ DL_PUSC_SLOT_SYMBOL_NUM) 
                                 * DL_PUSC_SLOT_SYMBOL_NUM;
    int allo_symbol = (int)ceil ( dl_relay_access_zone_bwa_free / (tmp_dl_relay_access_flow.size()-i));
    allo_symbol     = (int)ceil ( allo_symbol/ DL_PUSC_SLOT_SYMBOL_NUM) 
                             * DL_PUSC_SLOT_SYMBOL_NUM;
      
    if(allo_symbol_max <= allo_symbol + tmp_flow_queue->BWA_Total)
    {
      allo_symbol = allo_symbol_max - tmp_flow_queue->BWA_Total;
    }
    tmp_flow_queue->BWA_Total += allo_symbol;
      
    dl_relay_access_zone_bwa_free -= allo_symbol;
      
    if(dl_relay_access_zone_bwa_free <0)
    {
      lwx_err_msg("RR's DL RR", "dl_relay_access_zone_bwa_free <0");
    }

  }
  dl_relay_access_zone_bwa = dl_relay_access_zone_bwa_free;
  

  // final dl BWA setting
  for( int i=0; i <  tmp_dl_relay_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_dl_relay_access_flow.at(i);
    
    
    tmp_flow_queue->BWA_Total_Byte = (int)(tmp_flow_queue->BWA_Total 
                                     * tmp_flow_queue->BWA_Rate+ tmp_flow_queue->BWA_Backoff) ;
    tmp_flow_queue->BWA_Backoff = 0 ;
    
  }
  // log each phase's bandwidth allocation of each downlink relay flow
  ostringstream dl_relay_bwa_log;
  dl_relay_bwa_log << "============ dl relay bwa log: ============" << endl;
  dl_relay_bwa_log << "{(src_node_id/from_node_id/to_node_id) (src_ip/src_port/dst_port)\n"
                    << " (flow_type/rate) qsize (queue_size) \n"
                    << "BWA: (Total/Total_Byte/Flow_type/Next_Hop/Rate)} \n";
  
  for( int i=0; i <  tmp_dl_relay_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_dl_relay_flow.at(i);
    
    dl_relay_bwa_log  << "{("
                      << tmp_flow_queue->src_node_id << "/" 
                      << tmp_flow_queue->from_node_id << "/" 
                      << tmp_flow_queue->to_node_id << "/"  
                      << tmp_flow_queue->dst_port << ") " ;
    
    dl_relay_bwa_log  << "(" << tmp_flow_queue->src_ip << "/" 
                      << tmp_flow_queue->src_port << "/" 
                      << tmp_flow_queue->dst_ip << "/"  
                      << tmp_flow_queue->dst_port << ") ";
                      
    dl_relay_bwa_log  << "(" << tmp_flow_queue->flow_type << "/" 
                      << tmp_flow_queue->rate << ") ";
    
    dl_relay_bwa_log << "qsize (" << tmp_flow_queue->queue_size() << ") " ;
    
    dl_relay_bwa_log << "BWA: (" 
                      << tmp_flow_queue->BWA_Total << "/" 
                      << tmp_flow_queue->BWA_Total_Byte << "/" 
                      << tmp_flow_queue->BWA_Flow_Type << "/" 
                      << tmp_flow_queue->BWA_Next_Hop << "/" 
                      << tmp_flow_queue->BWA_Rate << ")} " ;
                      
  }
  // satisfy each dl relay flow by RR
  dl_relay_zone_bwa_free = dl_relay_zone_bwa;
  for( int i=0; i <  tmp_dl_relay_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_dl_relay_flow.at(i);
    
    int allo_symbol_max = (int)floor((tmp_flow_queue->Rmax)
                                 /(8*PACKETS_PER_SECOND*tmp_flow_queue->BWA_Rate));
    allo_symbol_max     = (int)ceil ( allo_symbol_max/ DL_PUSC_SLOT_SYMBOL_NUM) 
                                 * DL_PUSC_SLOT_SYMBOL_NUM;
    int allo_symbol = (int)ceil ( dl_relay_zone_bwa_free / (tmp_dl_relay_flow.size()-i));
    allo_symbol     = (int)ceil ( allo_symbol/ DL_PUSC_SLOT_SYMBOL_NUM) 
                             * DL_PUSC_SLOT_SYMBOL_NUM;
      
    if(allo_symbol_max <= allo_symbol + tmp_flow_queue->BWA_Total)
    {
      allo_symbol = allo_symbol_max - tmp_flow_queue->BWA_Total;
    }
    tmp_flow_queue->BWA_Total += allo_symbol;
      
    dl_relay_zone_bwa_free -= allo_symbol;
      
    if(dl_relay_zone_bwa_free <0)
    {
      lwx_err_msg("RR's DL Relay RR", "dl_relay_zone_bwa_free <0");
    }

  }
  dl_relay_zone_bwa = dl_relay_zone_bwa_free;
  

  // final dl BWA setting
  for( int i=0; i <  tmp_dl_relay_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_dl_relay_flow.at(i);
    
    
    tmp_flow_queue->BWA_Total_Byte = (int)(tmp_flow_queue->BWA_Total 
                                     * tmp_flow_queue->BWA_Rate + tmp_flow_queue->BWA_Backoff) ;
    tmp_flow_queue->BWA_Backoff = 0 ;
    
  }
  
  //
  //  uplink access bandwidth allocation
  //


  // satisfy each ul flow's Rmin
  ul_access_zone_bwa_free = ul_access_zone_bwa;
  for( int i=0; i <  tmp_ul_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ul_access_flow.at(i);
    
    if( tmp_flow_queue->QoS_Class == 1 || tmp_flow_queue->QoS_Class == 2 ||
        tmp_flow_queue->QoS_Class == 3 || tmp_flow_queue->QoS_Class == 4 )
    {
      int allo_symbol = (int)ceil((tmp_flow_queue->Rmin)
                              /(8*PACKETS_PER_SECOND*tmp_flow_queue->BWA_Rate));
      allo_symbol     = (int)ceil ( allo_symbol/ UL_PUSC_SLOT_SYMBOL_NUM) 
                                * UL_PUSC_SLOT_SYMBOL_NUM;
      
      if(allo_symbol > ul_access_zone_bwa_free)
      {
        allo_symbol=ul_access_zone_bwa_free; 
      }
      
      tmp_flow_queue->BWA.push_back(allo_symbol);
      tmp_flow_queue->BWA_Total+=allo_symbol;
      
      ul_access_zone_bwa_free -= allo_symbol;
      
      if(ul_access_zone_bwa_free <0)
      {
        lwx_err_msg("RR's UL Rmin", "ul_free_symbol_num <0");
      }
      
    }
  }
  ul_access_zone_bwa = ul_access_zone_bwa_free;
  
  
  // satisfy each ul flow by RR
  ul_access_zone_bwa_free = ul_access_zone_bwa;
  for( int i=0; i <  tmp_ul_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ul_access_flow.at(i);
    
    int allo_symbol_max = (int)floor((tmp_flow_queue->Rmax)
                                 /(8*PACKETS_PER_SECOND*tmp_flow_queue->BWA_Rate));
    allo_symbol_max     = (int)ceil ( allo_symbol_max/ UL_PUSC_SLOT_SYMBOL_NUM) 
                                 * UL_PUSC_SLOT_SYMBOL_NUM;
    int allo_symbol = (int)ceil ( ul_access_zone_bwa_free / (tmp_ul_access_flow.size()-i));
    allo_symbol     = (int)ceil ( allo_symbol/ UL_PUSC_SLOT_SYMBOL_NUM) 
                             * UL_PUSC_SLOT_SYMBOL_NUM;
      
    if(allo_symbol_max <= allo_symbol + tmp_flow_queue->BWA_Total)
    {
      allo_symbol = allo_symbol_max - tmp_flow_queue->BWA_Total;
    }
    tmp_flow_queue->BWA_Total += allo_symbol;
      
    ul_access_zone_bwa_free -= allo_symbol;
      
    if(ul_access_zone_bwa_free <0)
    {
      lwx_err_msg("RR's UL RR", "ul_access_zone_bwa_free <0");
    }

  }
  ul_access_zone_bwa = ul_access_zone_bwa_free;
  

  // final ul BWA setting
  for( int i=0; i <  tmp_ul_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ul_access_flow.at(i);
    
    
    tmp_flow_queue->BWA_Total_Byte = (int)(tmp_flow_queue->BWA_Total 
                                     * tmp_flow_queue->BWA_Rate + tmp_flow_queue->BWA_Backoff) ;
    tmp_flow_queue->BWA_Backoff = 0 ;
    
  }
  
  // log each phase's bandwidth allocation of each downlink flow
  ostringstream ul_access_bwa_log;
  ul_access_bwa_log << "============ ul access bwa log: ============" << endl;
  ul_access_bwa_log << "{(src_node_id/from_node_id/to_node_id) (src_ip/src_port/dst_port)\n"
                    << " (flow_type/rate) qsize (queue_size) \n"
                    << "BWA: (Total/Total_Byte/Flow_type/Next_Hop/Rate)} \n";
  
  for( int i=0; i <  tmp_ul_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ul_access_flow.at(i);
    
    ul_access_bwa_log << "{("
                      << tmp_flow_queue->src_node_id << "/" 
                      << tmp_flow_queue->from_node_id << "/" 
                      << tmp_flow_queue->to_node_id << "/"  
                      << tmp_flow_queue->dst_port << ") " ;
    
    ul_access_bwa_log << "(" << tmp_flow_queue->src_ip << "/" 
                      << tmp_flow_queue->src_port << "/" 
                      << tmp_flow_queue->dst_ip << "/"  
                      << tmp_flow_queue->dst_port << ") ";
                      
    ul_access_bwa_log << "(" << tmp_flow_queue->flow_type << "/" 
                      << tmp_flow_queue->rate << ") ";
    
    ul_access_bwa_log << "qsize (" << tmp_flow_queue->queue_size() << ") " ;
    
    ul_access_bwa_log << "BWA: (" 
                      << tmp_flow_queue->BWA_Total << "/" 
                      << tmp_flow_queue->BWA_Total_Byte << "/" 
                      << tmp_flow_queue->BWA_Flow_Type << "/" 
                      << tmp_flow_queue->BWA_Next_Hop << "/" 
                      << tmp_flow_queue->BWA_Rate << ")} " ;
                      
  }

  //
  //  uplink relay access bandwidth allocation
  //


  // satisfy each ul flow's Rmin
  ul_relay_access_zone_bwa_free = ul_relay_access_zone_bwa;
  for( int i=0; i <  tmp_ul_relay_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ul_relay_access_flow.at(i);
    
    if( tmp_flow_queue->QoS_Class == 1 || tmp_flow_queue->QoS_Class == 2 ||
        tmp_flow_queue->QoS_Class == 3 || tmp_flow_queue->QoS_Class == 4 )
    {
      int allo_symbol = (int)ceil((tmp_flow_queue->Rmin)
                              /(8*PACKETS_PER_SECOND*tmp_flow_queue->BWA_Rate));
      allo_symbol     = (int)ceil ( allo_symbol/ UL_PUSC_SLOT_SYMBOL_NUM) 
                                * UL_PUSC_SLOT_SYMBOL_NUM;
      
      if(allo_symbol > ul_relay_access_zone_bwa_free)
      {
        allo_symbol=ul_relay_access_zone_bwa_free; 
      }
      
      tmp_flow_queue->BWA.push_back(allo_symbol);
      tmp_flow_queue->BWA_Total+=allo_symbol;
      
      ul_relay_access_zone_bwa_free -= allo_symbol;
      
      if(ul_relay_access_zone_bwa_free <0)
      {
        lwx_err_msg("RR's UL Rmin", "ul_free_symbol_num <0");
      }
      
    }
  }
  ul_relay_access_zone_bwa = ul_relay_access_zone_bwa_free;
  
  
  // satisfy each ul flow by RR
  ul_relay_access_zone_bwa_free = ul_relay_access_zone_bwa;
  for( int i=0; i <  tmp_ul_relay_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ul_relay_access_flow.at(i);
    
    int allo_symbol_max = (int)floor((tmp_flow_queue->Rmax)
                                 /(8*PACKETS_PER_SECOND*tmp_flow_queue->BWA_Rate));
    allo_symbol_max     = (int)ceil ( allo_symbol_max/ UL_PUSC_SLOT_SYMBOL_NUM) 
                                 * UL_PUSC_SLOT_SYMBOL_NUM;
    int allo_symbol = (int)ceil ( ul_relay_access_zone_bwa_free / (tmp_ul_relay_access_flow.size()-i));
    allo_symbol     = (int)ceil ( allo_symbol/ UL_PUSC_SLOT_SYMBOL_NUM) 
                             * UL_PUSC_SLOT_SYMBOL_NUM;
      
    if(allo_symbol_max <= allo_symbol + tmp_flow_queue->BWA_Total)
    {
      allo_symbol = allo_symbol_max - tmp_flow_queue->BWA_Total;
    }
    tmp_flow_queue->BWA_Total += allo_symbol;
      
    ul_relay_access_zone_bwa_free -= allo_symbol;
      
    if(ul_relay_access_zone_bwa_free <0)
    {
      lwx_err_msg("RR's UL RR", "ul_relay_access_zone_bwa_free <0");
    }

  }
  ul_relay_access_zone_bwa = ul_relay_access_zone_bwa_free;
  

  // final ul BWA setting
  for( int i=0; i <  tmp_ul_relay_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ul_relay_access_flow.at(i);
    
    
    tmp_flow_queue->BWA_Total_Byte = (int)(tmp_flow_queue->BWA_Total 
                                     * tmp_flow_queue->BWA_Rate + tmp_flow_queue->BWA_Backoff) ;
    tmp_flow_queue->BWA_Backoff = 0 ;
    
  }

  // satisfy each ul relay flow by RR
  ul_relay_zone_bwa_free = ul_relay_zone_bwa;
  for( int i=0; i <  tmp_ul_relay_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ul_relay_flow.at(i);
    
    int allo_symbol_max = (int)floor((tmp_flow_queue->Rmax)
                                 /(8*PACKETS_PER_SECOND*tmp_flow_queue->BWA_Rate));
    allo_symbol_max     = (int)ceil ( allo_symbol_max/ UL_PUSC_SLOT_SYMBOL_NUM) 
                                 * UL_PUSC_SLOT_SYMBOL_NUM;
    int allo_symbol = (int)ceil ( ul_relay_zone_bwa_free / (tmp_ul_relay_flow.size()-i));
    allo_symbol     = (int)ceil ( allo_symbol/ UL_PUSC_SLOT_SYMBOL_NUM) 
                             * UL_PUSC_SLOT_SYMBOL_NUM;
      
    if(allo_symbol_max <= allo_symbol + tmp_flow_queue->BWA_Total)
    {
      allo_symbol = allo_symbol_max - tmp_flow_queue->BWA_Total;
    }
    tmp_flow_queue->BWA_Total += allo_symbol;
      
    ul_relay_zone_bwa_free -= allo_symbol;
      
    if(ul_relay_zone_bwa_free <0)
    {
      lwx_err_msg("RR's UL Relay RR", "ul_relay_zone_bwa_free <0");
    }

  }
  ul_relay_zone_bwa = ul_relay_zone_bwa_free;
  

  // final ul BWA setting
  for( int i=0; i <  tmp_ul_relay_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ul_relay_flow.at(i);
    
    
    tmp_flow_queue->BWA_Total_Byte = (int)(tmp_flow_queue->BWA_Total 
                                     * tmp_flow_queue->BWA_Rate + tmp_flow_queue->BWA_Backoff) ;
    tmp_flow_queue->BWA_Backoff = 0 ;
    
  }
  
// log each phase's bandwidth allocation of each downlink relay flow
  ostringstream ul_relay_bwa_log_odd;
  ul_relay_bwa_log_odd << "============ ul relay bwa log: ============" << endl;
  ul_relay_bwa_log_odd << "{(src_node_id/from_node_id/to_node_id) (src_ip/src_port/dst_port)\n"
                    << " (flow_type/rate) qsize (queue_size) \n"
                    << "BWA: (Total/Total_Byte/Flow_type/Next_Hop/Rate)} \n";
  
  for( int i=0; i <  tmp_ul_relay_flow_odd.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ul_relay_flow_odd.at(i);
    
    ul_relay_bwa_log_odd  << "{("
                      << tmp_flow_queue->src_node_id << "/" 
                      << tmp_flow_queue->from_node_id << "/" 
                      << tmp_flow_queue->to_node_id << "/"  
                      << tmp_flow_queue->dst_port << ") " ;
    
    ul_relay_bwa_log_odd  << "(" << tmp_flow_queue->src_ip << "/" 
                      << tmp_flow_queue->src_port << "/" 
                      << tmp_flow_queue->dst_ip << "/"  
                      << tmp_flow_queue->dst_port << ") ";
                      
    ul_relay_bwa_log_odd  << "(" << tmp_flow_queue->flow_type << "/" 
                      << tmp_flow_queue->rate << ") ";
    
    ul_relay_bwa_log_odd << "qsize (" << tmp_flow_queue->queue_size() << ") " ;
    
    ul_relay_bwa_log_odd << "BWA: (" 
                      << tmp_flow_queue->BWA_Total << "/" 
                      << tmp_flow_queue->BWA_Total_Byte << "/" 
                      << tmp_flow_queue->BWA_Flow_Type << "/" 
                      << tmp_flow_queue->BWA_Next_Hop << "/" 
                      << tmp_flow_queue->BWA_Rate << ")} " ;

  /*****************************
  * END OF PANDELI'S SINGLE FRAME ALGORITHM *
  *****************************/
}

void LWX::Multi_Frame(int node_id)
{
  /******************************
  * PANDELI'S ALGORITHM   multi-frame      *
  ******************************/

  // create pointer to each flow type's queue to allocation bandwidth efficiencially
  vector <Flow_Queue*> tmp_dl_access_flow;
  vector <Flow_Queue*> tmp_dl_relay_access_flow;
  vector <Flow_Queue*> tmp_dl_relay_flow;
  vector <Flow_Queue*> tmp_ul_access_flow;
  vector <Flow_Queue*> tmp_ul_relay_access_flow;
  vector <Flow_Queue*> tmp_ul_relay_flow;
  vector <Flow_Queue*> tmp_dl_relay_flow_odd;
  vector <Flow_Queue*> tmp_ul_relay_flow_odd;
  
  // get symbol num of DL and UL
  int dl_access_zone_bwa  = DL_ACCESS_ZONE_SYMBOL * DL_PUSC_SUBCH_NUM / AZ;
  int dl_relay_access_zone_bwa  = DL_ACCESS_ZONE_SYMBOL * DL_PUSC_SUBCH_NUM * NH /AZ;
  int dl_relay_zone_bwa   = DL_RELAY_ZONE_SYMBOL * DL_PUSC_SUBCH_NUM  / Nb_RS; 
  int ul_access_zone_bwa  = UL_ACCESS_ZONE_SYMBOL * UL_PUSC_SUBCH_NUM / AZ;
  int ul_relay_access_zone_bwa  = UL_ACCESS_ZONE_SYMBOL * UL_PUSC_SUBCH_NUM * NH / AZ;
  int ul_relay_zone_bwa   = UL_RELAY_ZONE_SYMBOL * UL_PUSC_SUBCH_NUM  / Nb_RS; 
  int dl_relay_zone_bwa_odd   = DL_RELAY_ZONE_SYMBOL * DL_PUSC_SUBCH_NUM  / Nb_RS; 
  int ul_relay_zone_bwa_odd   = UL_RELAY_ZONE_SYMBOL * UL_PUSC_SUBCH_NUM  / Nb_RS; 
  
  // algorithm calculation var
  int dl_access_zone_bwa_free=0;
  int dl_relay_access_zone_bwa_free=0;
  int dl_relay_zone_bwa_free=0;
  int ul_access_zone_bwa_free=0;
  int ul_relay_access_zone_bwa_free=0;
  int ul_relay_zone_bwa_free=0;
  int dl_relay_zone_bwa_free_odd=0;
  int ul_relay_zone_bwa_free_odd=0;

  
  // reset all relative var's value in flow's attribute
  for(int i=0; i < lwx_node_list.size(); i++)
  {
    LWX_Node *tmp_node = &lwx_node_list.at(i);
    for( int j=0; j <  tmp_node->ctrl_flow.size(); j++)
    { 
      
      Flow_Queue *tmp_flow_queue = &tmp_node->ctrl_flow.at(j);
      tmp_flow_queue->BWA.clear();
      tmp_flow_queue->BWA_Total=-1;
      tmp_flow_queue->BWA_Rate=-1;
      tmp_flow_queue->BWA_Total_Byte=-1;
      tmp_flow_queue->BWA_Flow_Type=-1;
      tmp_flow_queue->BWA_Next_Hop=-1;
    }
  }



  // =========================================================
  // path selection:
  //   1.find the path (directly to bs or passing through rs)
  //   2.set BWA_Flow_Type and BWA_Next_Hop
  // =========================================================
  
  // set relay route's access link (second hop) to its own flow_type
  // Set access zone flow for ss, flow_type is the default flow type
  // e.g. rs->ss_tp 
  for(int i=0; i < lwx_node_list.size(); i++)
  {
    LWX_Node *tmp_node = &lwx_node_list.at(i);
    for( int j=0; j <  tmp_node->ctrl_flow.size(); j++)
    { 
      Flow_Queue *tmp_flow_queue = &tmp_node->ctrl_flow.at(j);
      switch(tmp_flow_queue->flow_type)
      {
        case FLOW_TYPE_RS_TO_SS_TP:
          tmp_flow_queue->BWA_Flow_Type = tmp_flow_queue->flow_type;
          tmp_flow_queue->BWA_Next_Hop  = tmp_flow_queue->to_node_id;
          tmp_flow_queue->BWA_Rate      = tmp_flow_queue->rate;
          break;
        default:
          break;
      }
      
    }
  }
  
  // path selection:
  //   # if this flow has relay link, use the first relay link as its next hop
  //   # else use access link
  for(int i=0; i < lwx_node_list.size(); i++)
  {
    LWX_Node *tmp_node = &lwx_node_list.at(i);
    for( int j=0; j <  tmp_node->ctrl_flow.size(); j++)
    { 
      Flow_Queue *tmp_flow_queue = &tmp_node->ctrl_flow.at(j);
      if(tmp_flow_queue->flow_type == FLOW_TYPE_BS_TO_SS)
      {
        if(!tmp_flow_queue->relay_link.empty())
        {
          tmp_flow_queue->BWA_Flow_Type = tmp_flow_queue->relay_link.at(0).flow_type;
          tmp_flow_queue->BWA_Next_Hop  = tmp_flow_queue->relay_link.at(0).rs_node_id;
          tmp_flow_queue->BWA_Rate      = tmp_flow_queue->relay_link.at(0).rate;
        }
        else
        {
          tmp_flow_queue->BWA_Flow_Type = tmp_flow_queue->flow_type;
          tmp_flow_queue->BWA_Next_Hop  = tmp_flow_queue->to_node_id;
          tmp_flow_queue->BWA_Rate      = tmp_flow_queue->rate;
        }
      }
      
     if(tmp_flow_queue->flow_type == FLOW_TYPE_SS_TO_BS)
      {
        if(!tmp_flow_queue->relay_link.empty())
        {
          tmp_flow_queue->BWA_Flow_Type = tmp_flow_queue->relay_link.at(0).flow_type;
          tmp_flow_queue->BWA_Next_Hop  = tmp_flow_queue->relay_link.at(0).rs_node_id;
          tmp_flow_queue->BWA_Rate      = tmp_flow_queue->relay_link.at(0).rate;
        }
        else
        {
          tmp_flow_queue->BWA_Flow_Type = tmp_flow_queue->flow_type;
          tmp_flow_queue->BWA_Next_Hop  = tmp_flow_queue->to_node_id;
          tmp_flow_queue->BWA_Rate      = tmp_flow_queue->rate;
        }
      }
      
     if((tmp_flow_queue->flow_type == FLOW_TYPE_RS_TO_SS_NONTP)||
        (tmp_flow_queue->flow_type == FLOW_TYPE_RS_TO_BS))
      {
        if(!tmp_flow_queue->relay_link.empty())
        {
          tmp_flow_queue->BWA_Flow_Type = tmp_flow_queue->relay_link.at(0).flow_type;
          tmp_flow_queue->BWA_Next_Hop  = tmp_flow_queue->relay_link.at(0).rs_node_id;
          tmp_flow_queue->BWA_Rate      = tmp_flow_queue->relay_link.at(0).rate;
        }
        else
        {
          tmp_flow_queue->BWA_Flow_Type = tmp_flow_queue->flow_type;
          tmp_flow_queue->BWA_Next_Hop  = tmp_flow_queue->to_node_id;
          tmp_flow_queue->BWA_Rate      = tmp_flow_queue->rate;
          //lwx_err_msg("in relay zone", 
            // "wrong flow type, there is no connection for this relay zone DL BS- to RS");
        }
      }

   }
  }
  
  // =========================================================
  // bandwidth allocation:
  //   # access zone: RR + min QoS guarantee
  //   # relay zone:  only RR
  // =========================================================

  // collect flow into corresponding flow queue
  
  // !!!!!!! very important design: !!!!!!!!!!! Modified to wok for NT RS
  //   bs->ss and rs->ss should be scheduling together, because they use same bandwidth
  //   ss->bs and ss->rs should be also scheduling together, because they use same bandwidth
  
  for(int i=0; i < lwx_node_list.size(); i++)
  {
    LWX_Node *tmp_node = &lwx_node_list.at(i);
    for( int j=0; j <  tmp_node->ctrl_flow.size(); j++)
    { 
      switch(tmp_node->ctrl_flow.at(j).BWA_Flow_Type)
      {
        case FLOW_TYPE_BS_TO_SS:
          //tmp_dl_access_flow.push_back(&tmp_node->ctrl_flow.at(j));
        case FLOW_TYPE_RS_TO_SS_TP:
          tmp_dl_access_flow.push_back(&tmp_node->ctrl_flow.at(j));
         break;
        case FLOW_TYPE_RS_TO_SS_NONTP:
          tmp_dl_relay_access_flow.push_back(&tmp_node->ctrl_flow.at(j));
         // lwx_err_msg("OK BS to RS \n");
         break; // new
        case FLOW_TYPE_BS_TO_RS:
        case FLOW_TYPE_RS_TO_RS_NONTP_DL:
          tmp_dl_relay_flow.push_back(&tmp_node->ctrl_flow.at(j));
         break; 
        case FLOW_TYPE_SS_TO_BS:
         // tmp_ul_access_flow.push_back(&tmp_node->ctrl_flow.at(j));
        case FLOW_TYPE_SS_TO_RS_TP:
          tmp_ul_access_flow.push_back(&tmp_node->ctrl_flow.at(j));
         break;
        case FLOW_TYPE_SS_TO_RS_NONTP:           
          tmp_ul_relay_access_flow.push_back(&tmp_node->ctrl_flow.at(j));
         break; //new 
        case FLOW_TYPE_RS_TO_BS: 
        case FLOW_TYPE_RS_TO_RS_NONTP_UL:         
          tmp_ul_relay_flow.push_back(&tmp_node->ctrl_flow.at(j));
          break;       
        case FLOW_TYPE_RS_TO_RS_NONTP_DL_ODD:
          tmp_dl_relay_flow_odd.push_back(&tmp_node->ctrl_flow.at(j));
           //lwx_err_msg("OK RS to RS \n");
          break; 
        case FLOW_TYPE_RS_TO_RS_NONTP_UL_ODD:     
          tmp_ul_relay_flow_odd.push_back(&tmp_node->ctrl_flow.at(j));
         //lwx_err_msg("OK RS to SS \n");
          break;
        default:
          lwx_err_msg("in bandwidth allocation", 
             "wrong flow type, this algorithm doesn't support 802.16j transparent RS");
      }
      
    }
  }
  

  //
  //  downlink access bandwidth allocation
  //


  // satisfy each dl flow's Rmin
  dl_access_zone_bwa_free = dl_access_zone_bwa;
  for( int i=0; i <  tmp_dl_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_dl_access_flow.at(i);
    
    if( tmp_flow_queue->QoS_Class == 1 || tmp_flow_queue->QoS_Class == 2 ||
        tmp_flow_queue->QoS_Class == 3 || tmp_flow_queue->QoS_Class == 4 )
    {
      int allo_symbol = (int)ceil((tmp_flow_queue->Rmin)
                              /(8*PACKETS_PER_SECOND*tmp_flow_queue->BWA_Rate));
      allo_symbol     = (int)ceil ( allo_symbol/ DL_PUSC_SLOT_SYMBOL_NUM) 
                                * DL_PUSC_SLOT_SYMBOL_NUM;
      
      if(allo_symbol > dl_access_zone_bwa_free)
      {
        allo_symbol=dl_access_zone_bwa_free; 
      }
      
      tmp_flow_queue->BWA.push_back(allo_symbol);
      tmp_flow_queue->BWA_Total+=allo_symbol;
      
      dl_access_zone_bwa_free -= allo_symbol;
      
      if(dl_access_zone_bwa_free <0)
      {
        lwx_err_msg("RR's DL Rmin", "dl_free_symbol_num <0");
      }
      
    }
  }
  dl_access_zone_bwa = dl_access_zone_bwa_free;
  
  
  // satisfy each dl flow by RR
  dl_access_zone_bwa_free = dl_access_zone_bwa;
  for( int i=0; i <  tmp_dl_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_dl_access_flow.at(i);
    
    int allo_symbol_max = (int)floor((tmp_flow_queue->Rmax)
                                 /(8*PACKETS_PER_SECOND*tmp_flow_queue->BWA_Rate));
    allo_symbol_max     = (int)ceil ( allo_symbol_max/ DL_PUSC_SLOT_SYMBOL_NUM) 
                                 * DL_PUSC_SLOT_SYMBOL_NUM;
    int allo_symbol = (int)ceil ( dl_access_zone_bwa_free / (tmp_dl_access_flow.size()-i));
    allo_symbol     = (int)ceil ( allo_symbol/ DL_PUSC_SLOT_SYMBOL_NUM) 
                             * DL_PUSC_SLOT_SYMBOL_NUM;
      
    if(allo_symbol_max <= allo_symbol + tmp_flow_queue->BWA_Total)
    {
      allo_symbol = allo_symbol_max - tmp_flow_queue->BWA_Total;
    }
    tmp_flow_queue->BWA_Total += allo_symbol;
      
    dl_access_zone_bwa_free -= allo_symbol;
      
    if(dl_access_zone_bwa_free <0)
    {
      lwx_err_msg("RR's DL RR", "dl_access_zone_bwa_free <0");
    }

  }
  dl_access_zone_bwa = dl_access_zone_bwa_free;
  

  // final dl BWA setting
  for( int i=0; i <  tmp_dl_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_dl_access_flow.at(i);
    
    
    tmp_flow_queue->BWA_Total_Byte = (int)(tmp_flow_queue->BWA_Total 
                                     * tmp_flow_queue->BWA_Rate+ tmp_flow_queue->BWA_Backoff) ;
    tmp_flow_queue->BWA_Backoff = 0 ;
    
  }
  
//
  //  downlink relay_access bandwidth allocation
  //


  // satisfy each dl flow's Rmin
  dl_relay_access_zone_bwa_free = dl_relay_access_zone_bwa;
  for( int i=0; i <  tmp_dl_relay_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_dl_relay_access_flow.at(i);
    
    if( tmp_flow_queue->QoS_Class == 1 || tmp_flow_queue->QoS_Class == 2 ||
        tmp_flow_queue->QoS_Class == 3 || tmp_flow_queue->QoS_Class == 4 )
    {
      int allo_symbol = (int)ceil((tmp_flow_queue->Rmin)
                              /(8*PACKETS_PER_SECOND*tmp_flow_queue->BWA_Rate));
      allo_symbol     = (int)ceil ( allo_symbol/ DL_PUSC_SLOT_SYMBOL_NUM) 
                                * DL_PUSC_SLOT_SYMBOL_NUM;
      
      if(allo_symbol > dl_relay_access_zone_bwa_free)
      {
        allo_symbol=dl_relay_access_zone_bwa_free; 
      }
      
      tmp_flow_queue->BWA.push_back(allo_symbol);
      tmp_flow_queue->BWA_Total+=allo_symbol;
      
      dl_relay_access_zone_bwa_free -= allo_symbol;
      
      if(dl_relay_access_zone_bwa_free <0)
      {
        lwx_err_msg("RR's DL Rmin", "dl_free_symbol_num <0");
      }
      
    }
  }
  dl_relay_access_zone_bwa = dl_relay_access_zone_bwa_free;
  
  
  // satisfy each dl flow by RR
  dl_relay_access_zone_bwa_free = dl_relay_access_zone_bwa;
  for( int i=0; i <  tmp_dl_relay_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_dl_relay_access_flow.at(i);
    
    int allo_symbol_max = (int)floor((tmp_flow_queue->Rmax)
                                 /(8*PACKETS_PER_SECOND*tmp_flow_queue->BWA_Rate));
    allo_symbol_max     = (int)ceil ( allo_symbol_max/ DL_PUSC_SLOT_SYMBOL_NUM) 
                                 * DL_PUSC_SLOT_SYMBOL_NUM;
    int allo_symbol = (int)ceil ( dl_relay_access_zone_bwa_free / (tmp_dl_relay_access_flow.size()-i));
    allo_symbol     = (int)ceil ( allo_symbol/ DL_PUSC_SLOT_SYMBOL_NUM) 
                             * DL_PUSC_SLOT_SYMBOL_NUM;
      
    if(allo_symbol_max <= allo_symbol + tmp_flow_queue->BWA_Total)
    {
      allo_symbol = allo_symbol_max - tmp_flow_queue->BWA_Total;
    }
    tmp_flow_queue->BWA_Total += allo_symbol;
      
    dl_relay_access_zone_bwa_free -= allo_symbol;
      
    if(dl_relay_access_zone_bwa_free <0)
    {
      lwx_err_msg("RR's DL RR", "dl_relay_access_zone_bwa_free <0");
    }

  }
  dl_relay_access_zone_bwa = dl_relay_access_zone_bwa_free;
  

  // final dl BWA setting
  for( int i=0; i <  tmp_dl_relay_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_dl_relay_access_flow.at(i);
    
    
    tmp_flow_queue->BWA_Total_Byte = (int)(tmp_flow_queue->BWA_Total 
                                     * tmp_flow_queue->BWA_Rate+ tmp_flow_queue->BWA_Backoff) ;
    tmp_flow_queue->BWA_Backoff = 0 ;
    
  }
  
  // log each phase's bandwidth allocation of each downlink flow
  ostringstream dl_access_bwa_log;
  dl_access_bwa_log << "============ dl access bwa log: ============" << endl;
  dl_access_bwa_log << "{(src_node_id/from_node_id/to_node_id) (src_ip/src_port/dst_port)\n"
                    << " (flow_type/rate) qsize (queue_size) \n"
                    << "BWA: (Total/Total_Byte/Flow_type/Next_Hop/Rate)} \n";
  
  for( int i=0; i <  tmp_dl_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_dl_access_flow.at(i);
    
    dl_access_bwa_log << "{("
                      << tmp_flow_queue->src_node_id << "/" 
                      << tmp_flow_queue->from_node_id << "/" 
                      << tmp_flow_queue->to_node_id << "/"  
                      << tmp_flow_queue->dst_port << ") " ;
    
    dl_access_bwa_log << "(" << tmp_flow_queue->src_ip << "/" 
                      << tmp_flow_queue->src_port << "/" 
                      << tmp_flow_queue->dst_ip << "/"  
                      << tmp_flow_queue->dst_port << ") ";
                      
    dl_access_bwa_log << "(" << tmp_flow_queue->flow_type << "/" 
                      << tmp_flow_queue->rate << ") ";
    
    dl_access_bwa_log << "qsize (" << tmp_flow_queue->queue_size() << ") " ;
    
    dl_access_bwa_log << "BWA: (" 
                      << tmp_flow_queue->BWA_Total << "/" 
                      << tmp_flow_queue->BWA_Total_Byte << "/" 
                      << tmp_flow_queue->BWA_Flow_Type << "/" 
                      << tmp_flow_queue->BWA_Next_Hop << "/" 
                      << tmp_flow_queue->BWA_Rate << ")} " ;
                      
  }
  //lwx_test_msg("RR's DL ACCESS BWA\n", dl_access_bwa_log.str());

  //
  //  downlink relay bandwidth allocation (rs->ss)
  //

  
  // satisfy each dl relay flow by RR
  dl_relay_zone_bwa_free = dl_relay_zone_bwa;
  for( int i=0; i <  tmp_dl_relay_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_dl_relay_flow.at(i);
    
    int allo_symbol_max = (int)floor((tmp_flow_queue->Rmax)
                                 /(8*PACKETS_PER_SECOND*tmp_flow_queue->BWA_Rate));
    allo_symbol_max     = (int)ceil ( allo_symbol_max/ DL_PUSC_SLOT_SYMBOL_NUM) 
                                 * DL_PUSC_SLOT_SYMBOL_NUM;
    int allo_symbol = (int)ceil ( dl_relay_zone_bwa_free / (tmp_dl_relay_flow.size()-i));
    allo_symbol     = (int)ceil ( allo_symbol/ DL_PUSC_SLOT_SYMBOL_NUM) 
                             * DL_PUSC_SLOT_SYMBOL_NUM;
      
    if(allo_symbol_max <= allo_symbol + tmp_flow_queue->BWA_Total)
    {
      allo_symbol = allo_symbol_max - tmp_flow_queue->BWA_Total;
    }
    tmp_flow_queue->BWA_Total += allo_symbol;
      
    dl_relay_zone_bwa_free -= allo_symbol;
      
    if(dl_relay_zone_bwa_free <0)
    {
      lwx_err_msg("RR's DL Relay RR", "dl_relay_zone_bwa_free <0");
    }

  }
  dl_relay_zone_bwa = dl_relay_zone_bwa_free;
  

  // final dl BWA setting
  for( int i=0; i <  tmp_dl_relay_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_dl_relay_flow.at(i);
    
    
    tmp_flow_queue->BWA_Total_Byte = (int)(tmp_flow_queue->BWA_Total 
                                     * tmp_flow_queue->BWA_Rate + tmp_flow_queue->BWA_Backoff) ;
    tmp_flow_queue->BWA_Backoff = 0 ;
    
  }
  
  // log each phase's bandwidth allocation of each downlink relay flow
  ostringstream dl_relay_bwa_log;
  dl_relay_bwa_log << "============ dl relay bwa log: ============" << endl;
  dl_relay_bwa_log << "{(src_node_id/from_node_id/to_node_id) (src_ip/src_port/dst_port)\n"
                    << " (flow_type/rate) qsize (queue_size) \n"
                    << "BWA: (Total/Total_Byte/Flow_type/Next_Hop/Rate)} \n";
  
  for( int i=0; i <  tmp_dl_relay_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_dl_relay_flow.at(i);
    
    dl_relay_bwa_log  << "{("
                      << tmp_flow_queue->src_node_id << "/" 
                      << tmp_flow_queue->from_node_id << "/" 
                      << tmp_flow_queue->to_node_id << "/"  
                      << tmp_flow_queue->dst_port << ") " ;
    
    dl_relay_bwa_log  << "(" << tmp_flow_queue->src_ip << "/" 
                      << tmp_flow_queue->src_port << "/" 
                      << tmp_flow_queue->dst_ip << "/"  
                      << tmp_flow_queue->dst_port << ") ";
                      
    dl_relay_bwa_log  << "(" << tmp_flow_queue->flow_type << "/" 
                      << tmp_flow_queue->rate << ") ";
    
    dl_relay_bwa_log << "qsize (" << tmp_flow_queue->queue_size() << ") " ;
    
    dl_relay_bwa_log << "BWA: (" 
                      << tmp_flow_queue->BWA_Total << "/" 
                      << tmp_flow_queue->BWA_Total_Byte << "/" 
                      << tmp_flow_queue->BWA_Flow_Type << "/" 
                      << tmp_flow_queue->BWA_Next_Hop << "/" 
                      << tmp_flow_queue->BWA_Rate << ")} " ;
                      
  }
  //lwx_test_msg("RR's DL RELAY BWA\n", dl_relay_bwa_log.str());

  //
  //  downlink relay bandwidth allocation odd hop RS 
  //

  
  // satisfy each dl relay flow by RR
  dl_relay_zone_bwa_free_odd = dl_relay_zone_bwa_odd;
  for( int i=0; i <  tmp_dl_relay_flow_odd.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_dl_relay_flow_odd.at(i);
    
    int allo_symbol_max = (int)floor((tmp_flow_queue->Rmax)
                                 /(8*PACKETS_PER_SECOND*tmp_flow_queue->BWA_Rate));
    allo_symbol_max     = (int)ceil ( allo_symbol_max/ DL_PUSC_SLOT_SYMBOL_NUM) 
                                 * DL_PUSC_SLOT_SYMBOL_NUM;
    int allo_symbol = (int)ceil ( dl_relay_zone_bwa_free_odd / (tmp_dl_relay_flow_odd.size()-i));
    allo_symbol     = (int)ceil ( allo_symbol/ DL_PUSC_SLOT_SYMBOL_NUM) 
                             * DL_PUSC_SLOT_SYMBOL_NUM;
      
    if(allo_symbol_max <= allo_symbol + tmp_flow_queue->BWA_Total)
    {
      allo_symbol = allo_symbol_max - tmp_flow_queue->BWA_Total;
    }
    tmp_flow_queue->BWA_Total += allo_symbol;
      
    dl_relay_zone_bwa_free_odd -= allo_symbol;
      
    if(dl_relay_zone_bwa_free_odd <0)
    {
      lwx_err_msg("RR's DL Relay RR", "dl_relay_zone_bwa_free_odd <0");
    }

  }
  dl_relay_zone_bwa_odd = dl_relay_zone_bwa_free_odd;
  

  // final dl BWA setting
  for( int i=0; i <  tmp_dl_relay_flow_odd.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_dl_relay_flow_odd.at(i);
    
    
    tmp_flow_queue->BWA_Total_Byte = (int)(tmp_flow_queue->BWA_Total 
                                     * tmp_flow_queue->BWA_Rate + tmp_flow_queue->BWA_Backoff) ;
    tmp_flow_queue->BWA_Backoff = 0 ;
    
  }
  
  // log each phase's bandwidth allocation of each downlink relay flow
  ostringstream dl_relay_bwa_log_odd;
  dl_relay_bwa_log_odd << "============ dl relay bwa log: ============" << endl;
  dl_relay_bwa_log_odd << "{(src_node_id/from_node_id/to_node_id) (src_ip/src_port/dst_port)\n"
                    << " (flow_type/rate) qsize (queue_size) \n"
                    << "BWA: (Total/Total_Byte/Flow_type/Next_Hop/Rate)} \n";
  
  for( int i=0; i <  tmp_dl_relay_flow_odd.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_dl_relay_flow_odd.at(i);
    
    dl_relay_bwa_log_odd  << "{("
                      << tmp_flow_queue->src_node_id << "/" 
                      << tmp_flow_queue->from_node_id << "/" 
                      << tmp_flow_queue->to_node_id << "/"  
                      << tmp_flow_queue->dst_port << ") " ;
    
    dl_relay_bwa_log_odd  << "(" << tmp_flow_queue->src_ip << "/" 
                      << tmp_flow_queue->src_port << "/" 
                      << tmp_flow_queue->dst_ip << "/"  
                      << tmp_flow_queue->dst_port << ") ";
                      
    dl_relay_bwa_log_odd  << "(" << tmp_flow_queue->flow_type << "/" 
                      << tmp_flow_queue->rate << ") ";
    
    dl_relay_bwa_log_odd << "qsize (" << tmp_flow_queue->queue_size() << ") " ;
    
    dl_relay_bwa_log_odd << "BWA: (" 
                      << tmp_flow_queue->BWA_Total << "/" 
                      << tmp_flow_queue->BWA_Total_Byte << "/" 
                      << tmp_flow_queue->BWA_Flow_Type << "/" 
                      << tmp_flow_queue->BWA_Next_Hop << "/" 
                      << tmp_flow_queue->BWA_Rate << ")} " ;
                      
  }
  //lwx_test_msg("RR's DL RELAY BWA\n", dl_relay_bwa_log.str());
  //
  //  uplink access bandwidth allocation
  //


  // satisfy each ul flow's Rmin
  ul_access_zone_bwa_free = ul_access_zone_bwa;
  for( int i=0; i <  tmp_ul_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ul_access_flow.at(i);
    
    if( tmp_flow_queue->QoS_Class == 1 || tmp_flow_queue->QoS_Class == 2 ||
        tmp_flow_queue->QoS_Class == 3 || tmp_flow_queue->QoS_Class == 4 )
    {
      int allo_symbol = (int)ceil((tmp_flow_queue->Rmin)
                              /(8*PACKETS_PER_SECOND*tmp_flow_queue->BWA_Rate));
      allo_symbol     = (int)ceil ( allo_symbol/ UL_PUSC_SLOT_SYMBOL_NUM) 
                                * UL_PUSC_SLOT_SYMBOL_NUM;
      
      if(allo_symbol > ul_access_zone_bwa_free)
      {
        allo_symbol=ul_access_zone_bwa_free; 
      }
      
      tmp_flow_queue->BWA.push_back(allo_symbol);
      tmp_flow_queue->BWA_Total+=allo_symbol;
      
      ul_access_zone_bwa_free -= allo_symbol;
      
      if(ul_access_zone_bwa_free <0)
      {
        lwx_err_msg("RR's UL Rmin", "ul_free_symbol_num <0");
      }
      
    }
  }
  ul_access_zone_bwa = ul_access_zone_bwa_free;
  
  
  // satisfy each ul flow by RR
  ul_access_zone_bwa_free = ul_access_zone_bwa;
  for( int i=0; i <  tmp_ul_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ul_access_flow.at(i);
    
    int allo_symbol_max = (int)floor((tmp_flow_queue->Rmax)
                                 /(8*PACKETS_PER_SECOND*tmp_flow_queue->BWA_Rate));
    allo_symbol_max     = (int)ceil ( allo_symbol_max/ UL_PUSC_SLOT_SYMBOL_NUM) 
                                 * UL_PUSC_SLOT_SYMBOL_NUM;
    int allo_symbol = (int)ceil ( ul_access_zone_bwa_free / (tmp_ul_access_flow.size()-i));
    allo_symbol     = (int)ceil ( allo_symbol/ UL_PUSC_SLOT_SYMBOL_NUM) 
                             * UL_PUSC_SLOT_SYMBOL_NUM;
      
    if(allo_symbol_max <= allo_symbol + tmp_flow_queue->BWA_Total)
    {
      allo_symbol = allo_symbol_max - tmp_flow_queue->BWA_Total;
    }
    tmp_flow_queue->BWA_Total += allo_symbol;
      
    ul_access_zone_bwa_free -= allo_symbol;
      
    if(ul_access_zone_bwa_free <0)
    {
      lwx_err_msg("RR's UL RR", "ul_access_zone_bwa_free <0");
    }

  }
  ul_access_zone_bwa = ul_access_zone_bwa_free;
  

  // final ul BWA setting
  for( int i=0; i <  tmp_ul_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ul_access_flow.at(i);
    
    
    tmp_flow_queue->BWA_Total_Byte = (int)(tmp_flow_queue->BWA_Total 
                                     * tmp_flow_queue->BWA_Rate + tmp_flow_queue->BWA_Backoff) ;
    tmp_flow_queue->BWA_Backoff = 0 ;
    
  }
  
//
  //  uplink relay_access bandwidth allocation
  //


  // satisfy each ul flow's Rmin
  ul_relay_access_zone_bwa_free = ul_relay_access_zone_bwa;
  for( int i=0; i <  tmp_ul_relay_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ul_relay_access_flow.at(i);
    
    if( tmp_flow_queue->QoS_Class == 1 || tmp_flow_queue->QoS_Class == 2 ||
        tmp_flow_queue->QoS_Class == 3 || tmp_flow_queue->QoS_Class == 4 )
    {
      int allo_symbol = (int)ceil((tmp_flow_queue->Rmin)
                              /(8*PACKETS_PER_SECOND*tmp_flow_queue->BWA_Rate));
      allo_symbol     = (int)ceil ( allo_symbol/ UL_PUSC_SLOT_SYMBOL_NUM) 
                                * UL_PUSC_SLOT_SYMBOL_NUM;
      
      if(allo_symbol > ul_relay_access_zone_bwa_free)
      {
        allo_symbol=ul_relay_access_zone_bwa_free; 
      }
      
      tmp_flow_queue->BWA.push_back(allo_symbol);
      tmp_flow_queue->BWA_Total+=allo_symbol;
      
      ul_relay_access_zone_bwa_free -= allo_symbol;
      
      if(ul_relay_access_zone_bwa_free <0)
      {
        lwx_err_msg("RR's UL Rmin", "ul_free_symbol_num <0");
      }
      
    }
  }
  ul_relay_access_zone_bwa = ul_relay_access_zone_bwa_free;
  
  
  // satisfy each ul flow by RR
  ul_relay_access_zone_bwa_free = ul_relay_access_zone_bwa;
  for( int i=0; i <  tmp_ul_relay_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ul_relay_access_flow.at(i);
    
    int allo_symbol_max = (int)floor((tmp_flow_queue->Rmax)
                                 /(8*PACKETS_PER_SECOND*tmp_flow_queue->BWA_Rate));
    allo_symbol_max     = (int)ceil ( allo_symbol_max/ UL_PUSC_SLOT_SYMBOL_NUM) 
                                 * UL_PUSC_SLOT_SYMBOL_NUM;
    int allo_symbol = (int)ceil ( ul_relay_access_zone_bwa_free / (tmp_ul_relay_access_flow.size()-i));
    allo_symbol     = (int)ceil ( allo_symbol/ UL_PUSC_SLOT_SYMBOL_NUM) 
                             * UL_PUSC_SLOT_SYMBOL_NUM;
      
    if(allo_symbol_max <= allo_symbol + tmp_flow_queue->BWA_Total)
    {
      allo_symbol = allo_symbol_max - tmp_flow_queue->BWA_Total;
    }
    tmp_flow_queue->BWA_Total += allo_symbol;
      
    ul_relay_access_zone_bwa_free -= allo_symbol;
      
    if(ul_relay_access_zone_bwa_free <0)
    {
      lwx_err_msg("RR's UL RR", "ul_relay_access_zone_bwa_free <0");
    }

  }
  ul_relay_access_zone_bwa = ul_relay_access_zone_bwa_free;
  

  // final ul BWA setting
  for( int i=0; i <  tmp_ul_relay_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ul_relay_access_flow.at(i);
    
    
    tmp_flow_queue->BWA_Total_Byte = (int)(tmp_flow_queue->BWA_Total 
                                     * tmp_flow_queue->BWA_Rate + tmp_flow_queue->BWA_Backoff) ;
    tmp_flow_queue->BWA_Backoff = 0 ;
    
  }
  
  
  // log each phase's bandwidth allocation of each downlink flow
  ostringstream ul_access_bwa_log;
  ul_access_bwa_log << "============ ul access bwa log: ============" << endl;
  ul_access_bwa_log << "{(src_node_id/from_node_id/to_node_id) (src_ip/src_port/dst_port)\n"
                    << " (flow_type/rate) qsize (queue_size) \n"
                    << "BWA: (Total/Total_Byte/Flow_type/Next_Hop/Rate)} \n";
  
  for( int i=0; i <  tmp_ul_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ul_access_flow.at(i);
    
    ul_access_bwa_log << "{("
                      << tmp_flow_queue->src_node_id << "/" 
                      << tmp_flow_queue->from_node_id << "/" 
                      << tmp_flow_queue->to_node_id << "/"  
                      << tmp_flow_queue->dst_port << ") " ;
    
    ul_access_bwa_log << "(" << tmp_flow_queue->src_ip << "/" 
                      << tmp_flow_queue->src_port << "/" 
                      << tmp_flow_queue->dst_ip << "/"  
                      << tmp_flow_queue->dst_port << ") ";
                      
    ul_access_bwa_log << "(" << tmp_flow_queue->flow_type << "/" 
                      << tmp_flow_queue->rate << ") ";
    
    ul_access_bwa_log << "qsize (" << tmp_flow_queue->queue_size() << ") " ;
    
    ul_access_bwa_log << "BWA: (" 
                      << tmp_flow_queue->BWA_Total << "/" 
                      << tmp_flow_queue->BWA_Total_Byte << "/" 
                      << tmp_flow_queue->BWA_Flow_Type << "/" 
                      << tmp_flow_queue->BWA_Next_Hop << "/" 
                      << tmp_flow_queue->BWA_Rate << ")} " ;
                      
  }
  //lwx_test_msg("RR's UL ACCESS BWA\n", ul_access_bwa_log.str());

  //
  //  uplink relay bandwidth allocation (rs->ss) 
  //
  
  // satisfy each ul relay flow by RR
  ul_relay_zone_bwa_free = ul_relay_zone_bwa;
  for( int i=0; i <  tmp_ul_relay_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ul_relay_flow.at(i);
    
    int allo_symbol_max = (int)floor((tmp_flow_queue->Rmax)
                                 /(8*PACKETS_PER_SECOND*tmp_flow_queue->BWA_Rate));
    allo_symbol_max     = (int)ceil ( allo_symbol_max/ UL_PUSC_SLOT_SYMBOL_NUM) 
                                 * UL_PUSC_SLOT_SYMBOL_NUM;
    int allo_symbol = (int)ceil ( ul_relay_zone_bwa_free / (tmp_ul_relay_flow.size()-i));
    allo_symbol     = (int)ceil ( allo_symbol/ UL_PUSC_SLOT_SYMBOL_NUM) 
                             * UL_PUSC_SLOT_SYMBOL_NUM;
      
    if(allo_symbol_max <= allo_symbol + tmp_flow_queue->BWA_Total)
    {
      allo_symbol = allo_symbol_max - tmp_flow_queue->BWA_Total;
    }
    tmp_flow_queue->BWA_Total += allo_symbol;
      
    ul_relay_zone_bwa_free -= allo_symbol;
      
    if(ul_relay_zone_bwa_free <0)
    {
      lwx_err_msg("RR's UL Relay RR", "ul_relay_zone_bwa_free <0");
    }

  }
  ul_relay_zone_bwa = ul_relay_zone_bwa_free;
  

  // final ul BWA setting
  for( int i=0; i <  tmp_ul_relay_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ul_relay_flow.at(i);
    
    
    tmp_flow_queue->BWA_Total_Byte = (int)(tmp_flow_queue->BWA_Total 
                                     * tmp_flow_queue->BWA_Rate + tmp_flow_queue->BWA_Backoff) ;
    tmp_flow_queue->BWA_Backoff = 0 ;
    
  }
  
  // log each phase's bandwidth allocation of each downlink relay flow
  ostringstream ul_relay_bwa_log;
  ul_relay_bwa_log << "============ ul relay bwa log: ============" << endl;
  ul_relay_bwa_log << "{(src_node_id/from_node_id/to_node_id) (src_ip/src_port/dst_port)\n"
                    << " (flow_type/rate) qsize (queue_size) \n"
                    << "BWA: (Total/Total_Byte/Flow_type/Next_Hop/Rate)} \n";
  
  for( int i=0; i <  tmp_ul_relay_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ul_relay_flow.at(i);
    
    ul_relay_bwa_log  << "{("
                      << tmp_flow_queue->src_node_id << "/" 
                      << tmp_flow_queue->from_node_id << "/" 
                      << tmp_flow_queue->to_node_id << "/"  
                      << tmp_flow_queue->dst_port << ") " ;
    
    ul_relay_bwa_log  << "(" << tmp_flow_queue->src_ip << "/" 
                      << tmp_flow_queue->src_port << "/" 
                      << tmp_flow_queue->dst_ip << "/"  
                      << tmp_flow_queue->dst_port << ") ";
                      
    ul_relay_bwa_log  << "(" << tmp_flow_queue->flow_type << "/" 
                      << tmp_flow_queue->rate << ") ";
    
    ul_relay_bwa_log << "qsize (" << tmp_flow_queue->queue_size() << ") " ;
    
    ul_relay_bwa_log << "BWA: (" 
                      << tmp_flow_queue->BWA_Total << "/" 
                      << tmp_flow_queue->BWA_Total_Byte << "/" 
                      << tmp_flow_queue->BWA_Flow_Type << "/" 
                      << tmp_flow_queue->BWA_Next_Hop << "/" 
                      << tmp_flow_queue->BWA_Rate << ")} " ;
                      
  }
  //lwx_test_msg("RR's UL RELAY BWA\n", ul_relay_bwa_log.str());

  //
  //  uplink relay bandwidth allocation odd hop 
  //
  
  // satisfy each ul relay flow by RR
  ul_relay_zone_bwa_free_odd = ul_relay_zone_bwa_odd;
  for( int i=0; i <  tmp_ul_relay_flow_odd.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ul_relay_flow_odd.at(i);
    
    int allo_symbol_max = (int)floor((tmp_flow_queue->Rmax)
                                 /(8*PACKETS_PER_SECOND*tmp_flow_queue->BWA_Rate));
    allo_symbol_max     = (int)ceil ( allo_symbol_max/ UL_PUSC_SLOT_SYMBOL_NUM) 
                                 * UL_PUSC_SLOT_SYMBOL_NUM;
    int allo_symbol = (int)ceil ( ul_relay_zone_bwa_free_odd / (tmp_ul_relay_flow_odd.size()-i));
    allo_symbol     = (int)ceil ( allo_symbol/ UL_PUSC_SLOT_SYMBOL_NUM) 
                             * UL_PUSC_SLOT_SYMBOL_NUM;
      
    if(allo_symbol_max <= allo_symbol + tmp_flow_queue->BWA_Total)
    {
      allo_symbol = allo_symbol_max - tmp_flow_queue->BWA_Total;
    }
    tmp_flow_queue->BWA_Total += allo_symbol;
      
    ul_relay_zone_bwa_free_odd -= allo_symbol;
      
    if(ul_relay_zone_bwa_free_odd <0)
    {
      lwx_err_msg("RR's UL Relay RR", "ul_relay_zone_bwa_free_odd <0");
    }

  }
  ul_relay_zone_bwa_odd = ul_relay_zone_bwa_free_odd;
  

  // final ul BWA setting
  for( int i=0; i <  tmp_ul_relay_flow_odd.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ul_relay_flow_odd.at(i);
    
    
    tmp_flow_queue->BWA_Total_Byte = (int)(tmp_flow_queue->BWA_Total 
                                     * tmp_flow_queue->BWA_Rate + tmp_flow_queue->BWA_Backoff) ;
    tmp_flow_queue->BWA_Backoff = 0 ;
    
  }
  
  // log each phase's bandwidth allocation of each downlink relay flow
  ostringstream ul_relay_bwa_log_odd;
  ul_relay_bwa_log_odd << "============ ul relay bwa log: ============" << endl;
  ul_relay_bwa_log_odd << "{(src_node_id/from_node_id/to_node_id) (src_ip/src_port/dst_port)\n"
                    << " (flow_type/rate) qsize (queue_size) \n"
                    << "BWA: (Total/Total_Byte/Flow_type/Next_Hop/Rate)} \n";
  
  for( int i=0; i <  tmp_ul_relay_flow_odd.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ul_relay_flow_odd.at(i);
    
    ul_relay_bwa_log_odd  << "{("
                      << tmp_flow_queue->src_node_id << "/" 
                      << tmp_flow_queue->from_node_id << "/" 
                      << tmp_flow_queue->to_node_id << "/"  
                      << tmp_flow_queue->dst_port << ") " ;
    
    ul_relay_bwa_log_odd  << "(" << tmp_flow_queue->src_ip << "/" 
                      << tmp_flow_queue->src_port << "/" 
                      << tmp_flow_queue->dst_ip << "/"  
                      << tmp_flow_queue->dst_port << ") ";
                      
    ul_relay_bwa_log_odd  << "(" << tmp_flow_queue->flow_type << "/" 
                      << tmp_flow_queue->rate << ") ";
    
    ul_relay_bwa_log_odd << "qsize (" << tmp_flow_queue->queue_size() << ") " ;
    
    ul_relay_bwa_log_odd << "BWA: (" 
                      << tmp_flow_queue->BWA_Total << "/" 
                      << tmp_flow_queue->BWA_Total_Byte << "/" 
                      << tmp_flow_queue->BWA_Flow_Type << "/" 
                      << tmp_flow_queue->BWA_Next_Hop << "/" 
                      << tmp_flow_queue->BWA_Rate << ")} " ;
                      
  }
  //lwx_test_msg("RR's UL RELAY BWA\n", ul_relay_bwa_log.str());
  /*****************************
  * END OF PANDELI'S MF ALGORITHM *
  *****************************/
}

void LWX::BWA_Round_Robin(int node_id)
{

  // create pointer to downlink & uplink queue to allocation bandwidth efficiencially
  vector <Flow_Queue*> tmp_ss_to_bs;
  vector <Flow_Queue*> tmp_bs_to_ss;
  
  // get symbol num of DL and UL
  int dl_total_symbol_num = DL_SYMBOL_NUM * DL_PUSC_SUBCH_NUM ;
  int ul_total_symbol_num = UL_SYMBOL_NUM * UL_PUSC_SUBCH_NUM ;
  
  // algorithm calculation var
  int dl_free_symbol_num=0;
  int ul_free_symbol_num=0;
  
  // reset all relative var's value in flow's attribute
  for(int i=0; i < lwx_node_list.size(); i++)
  {
    LWX_Node *tmp_node = &lwx_node_list.at(i);
    for( int j=0; j <  tmp_node->ctrl_flow.size(); j++)
    { 
      
      Flow_Queue *tmp_flow_queue = &tmp_node->ctrl_flow.at(j);
      tmp_flow_queue->BWA.clear();
      tmp_flow_queue->BWA_Total=-1;
      tmp_flow_queue->BWA_Total_Byte=-1;
      tmp_flow_queue->BWA_Flow_Type=-1;
      tmp_flow_queue->BWA_Next_Hop=-1;
    }
  }
  
  // collect downlink and flow into corresponding flow queue
  for(int i=0; i < lwx_node_list.size(); i++)
  {
    LWX_Node *tmp_node = &lwx_node_list.at(i);
    for( int j=0; j <  tmp_node->ctrl_flow.size(); j++)
    { 
      if(tmp_node->ctrl_flow.at(j).flow_type==FLOW_TYPE_BS_TO_SS) // bs->ss
      {
        tmp_bs_to_ss.push_back(&tmp_node->ctrl_flow.at(j));
      }else if(tmp_node->ctrl_flow.at(j).flow_type==FLOW_TYPE_SS_TO_BS) // ss->bs
      {
        tmp_ss_to_bs.push_back(&tmp_node->ctrl_flow.at(j));
      }else
      {
        lwx_err_msg("in bandwidth allocation", 
           "wrong flow type, this algorithm doesn't support 802.16j");
      }
    }
  }
  
  // test log
  
  for(int i=0 ;i <tmp_bs_to_ss.size();i++)
  {
   
    ostringstream tmp_pkt_leng;
    tmp_pkt_leng << "dl in " << node_id << " "
                     << "(" << tmp_bs_to_ss.at(i)->flow_type << "/"
                     << tmp_bs_to_ss.at(i)->src_node_id << "/"
                     << tmp_bs_to_ss.at(i)->from_node_id << "/"
                     << tmp_bs_to_ss.at(i)->to_node_id << "/"
                     << tmp_bs_to_ss.at(i)->src_ip << "/"
                     << tmp_bs_to_ss.at(i)->src_port << "/"
                     << tmp_bs_to_ss.at(i)->dst_ip << "/"
                     << tmp_bs_to_ss.at(i)->dst_port << "/"
                     << tmp_bs_to_ss.at(i)->queue_length() << ")" << endl;
    //lwx_test_msg(tmp_pkt_leng.str());

  }

  for(int i=0 ;i <tmp_ss_to_bs.size();i++)
  {
   
    ostringstream tmp_pkt_leng;
    tmp_pkt_leng << "up in " << node_id << " "
                     << "(" << tmp_ss_to_bs.at(i)->flow_type << "/"
                     << tmp_ss_to_bs.at(i)->src_node_id << "/"
                     << tmp_ss_to_bs.at(i)->to_node_id << "/"
                     << tmp_ss_to_bs.at(i)->src_ip << "/"
                     << tmp_ss_to_bs.at(i)->src_port << "/"
                     << tmp_ss_to_bs.at(i)->dst_ip << "/"
                     << tmp_ss_to_bs.at(i)->dst_port << "/"
                     << tmp_ss_to_bs.at(i)->queue_length() << ")" << endl;
    //lwx_test_msg(tmp_pkt_leng.str());

  }
  
  //
  //  downlink bandwidth allocation
  //

  
  // satisfy each dl flow by RR
  dl_free_symbol_num = dl_total_symbol_num;
  for( int i=0; i <  tmp_bs_to_ss.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_bs_to_ss.at(i);
    
    int allo_symbol_max = (int)floor((tmp_flow_queue->Rmax)
                                 /(8*PACKETS_PER_SECOND*tmp_flow_queue->rate));
    allo_symbol_max     = (int)ceil ( allo_symbol_max/ DL_PUSC_SLOT_SYMBOL_NUM) 
                                 * DL_PUSC_SLOT_SYMBOL_NUM;
    int allo_symbol = (int)ceil ( dl_free_symbol_num / (tmp_bs_to_ss.size()-i));
    allo_symbol     = (int)ceil ( allo_symbol/ DL_PUSC_SLOT_SYMBOL_NUM) 
                             * DL_PUSC_SLOT_SYMBOL_NUM;
      
    if(allo_symbol_max <= allo_symbol + tmp_flow_queue->BWA_Total)
    {
      allo_symbol = allo_symbol_max - tmp_flow_queue->BWA_Total;
    }
    tmp_flow_queue->BWA_Total += allo_symbol;
      
    dl_free_symbol_num -= allo_symbol;
      
    if(dl_free_symbol_num <0)
    {
      lwx_err_msg("RR's DL RR", "dl_free_symbol_num <0");
    }

  }
  dl_total_symbol_num = dl_free_symbol_num;
  

  // final dl BWA setting
  for( int i=0; i <  tmp_bs_to_ss.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_bs_to_ss.at(i);
    
    
    tmp_flow_queue->BWA_Total_Byte = (int)(tmp_flow_queue->BWA_Total 
                                     * tmp_flow_queue->rate + tmp_flow_queue->BWA_Backoff) ;
    tmp_flow_queue->BWA_Backoff = 0 ;
    tmp_flow_queue->BWA_Flow_Type = FLOW_TYPE_BS_TO_SS ;
    
  }
  
  
  // log each phase's bandwidth allocation of each downlink flow
  ostringstream dl_bwa_log;
  dl_bwa_log.clear();
  for( int i=0; i <  tmp_bs_to_ss.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_bs_to_ss.at(i);
    
    dl_bwa_log << tmp_flow_queue->src_ip << "/" 
               << tmp_flow_queue->src_port << "/" 
               << tmp_flow_queue->dst_ip << "/"  
               << tmp_flow_queue->dst_port << " " << endl;
    
    dl_bwa_log << "queue_size: " << tmp_flow_queue->queue_size() << endl;
    
    for(int j=0; j < tmp_flow_queue->BWA.size(); j++)
    { 
      dl_bwa_log << tmp_flow_queue->BWA.at(j) <<  "/" ;
    }
    dl_bwa_log << tmp_flow_queue->BWA_Total << "\n" ;
  }
  //lwx_test_msg("RR's DL BWA\n", dl_bwa_log.str());


  //
  //  uplink bandwidth allocation
  //

  // satisfy each ul flow by RR
  ul_free_symbol_num = ul_total_symbol_num;
  for( int i=0; i <  tmp_ss_to_bs.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ss_to_bs.at(i);
    
    int allo_symbol_max = (int)floor((tmp_flow_queue->Rmax)
                                 /(8*PACKETS_PER_SECOND*tmp_flow_queue->rate));
    allo_symbol_max     = (int)ceil ( allo_symbol_max/ UL_PUSC_SLOT_SYMBOL_NUM) 
                                 * UL_PUSC_SLOT_SYMBOL_NUM;
    int allo_symbol = (int)ceil ( ul_free_symbol_num / (tmp_ss_to_bs.size()-i));
    allo_symbol     = (int)ceil ( allo_symbol/ UL_PUSC_SLOT_SYMBOL_NUM) 
                             * UL_PUSC_SLOT_SYMBOL_NUM;
      
    if(allo_symbol_max <= allo_symbol + tmp_flow_queue->BWA_Total)
    {
      allo_symbol = allo_symbol_max - tmp_flow_queue->BWA_Total;
    }
    tmp_flow_queue->BWA_Total += allo_symbol;
      
    ul_free_symbol_num -= allo_symbol;
      
    if(ul_free_symbol_num <0)
    {
      lwx_err_msg("RR's UL RR", "ul_free_symbol_num <0");
    }

  }
  ul_total_symbol_num = ul_free_symbol_num;
  
  // final ul BWA setting
  for( int i=0; i <  tmp_ss_to_bs.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ss_to_bs.at(i);

    tmp_flow_queue->BWA_Total_Byte = (int)(tmp_flow_queue->BWA_Total * tmp_flow_queue
                                            ->rate + tmp_flow_queue->BWA_Backoff) ;
    tmp_flow_queue->BWA_Backoff = 0 ;
    tmp_flow_queue->BWA_Flow_Type=FLOW_TYPE_SS_TO_BS;
  }
  
  
  // log each phase's bandwidth allocation of each uplink flow
  ostringstream ul_bwa_log;
  ul_bwa_log.clear();
  for( int i=0; i <  tmp_ss_to_bs.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ss_to_bs.at(i);
    
    ul_bwa_log << tmp_flow_queue->src_ip << "/" 
               << tmp_flow_queue->src_port << "/" 
               << tmp_flow_queue->dst_ip << "/"  
               << tmp_flow_queue->dst_port << " " << endl;
  
    ul_bwa_log << "queue_size: " << tmp_flow_queue->queue_size() << endl;
               
    for(int j=0; j < tmp_flow_queue->BWA.size(); j++)
    { 
      ul_bwa_log << tmp_flow_queue->BWA.at(j) <<  "/" ;
    }
    
    ul_bwa_log << tmp_flow_queue->BWA_Total << "\n" ;
  }
  //lwx_test_msg("RR's UL BWA\n", ul_bwa_log.str());
  
}

void LWX::BWA_Strict_Priority(int node_id)
{
  
  // create pointer to downlink & uplink queue to allocation bandwidth efficiencially
  vector <Flow_Queue*> tmp_ss_to_bs;
  vector <Flow_Queue*> tmp_bs_to_ss;
  
  // get symbol num of DL and UL
  int dl_total_symbol_num = DL_SYMBOL_NUM * DL_PUSC_SUBCH_NUM ;
  int ul_total_symbol_num = UL_SYMBOL_NUM * UL_PUSC_SUBCH_NUM ;
  
  // algorithm calculation var
  int dl_free_symbol_num=0;
  int ul_free_symbol_num=0;
  
  // reset all relative var's value in flow's attribute
  for(int i=0; i < lwx_node_list.size(); i++)
  {
    LWX_Node *tmp_node = &lwx_node_list.at(i);
    for( int j=0; j <  tmp_node->ctrl_flow.size(); j++)
    { 
      
      Flow_Queue *tmp_flow_queue = &tmp_node->ctrl_flow.at(j);
      tmp_flow_queue->BWA.clear();
      tmp_flow_queue->BWA_Total=-1;
      tmp_flow_queue->BWA_Total_Byte=-1;
      tmp_flow_queue->BWA_Flow_Type=-1;
      tmp_flow_queue->BWA_Next_Hop=-1;
    }
  }
  
  // collect downlink and flow into corresponding flow queue
  for(int i=0; i < lwx_node_list.size(); i++)
  {
    LWX_Node *tmp_node = &lwx_node_list.at(i);
    for( int j=0; j <  tmp_node->ctrl_flow.size(); j++)
    { 
      if(tmp_node->ctrl_flow.at(j).flow_type==FLOW_TYPE_BS_TO_SS) // bs->ss
      {
        tmp_bs_to_ss.push_back(&tmp_node->ctrl_flow.at(j));
      }else if(tmp_node->ctrl_flow.at(j).flow_type==FLOW_TYPE_SS_TO_BS) // ss->bs
      {
        tmp_ss_to_bs.push_back(&tmp_node->ctrl_flow.at(j));
      }else
      {
        lwx_err_msg("in bandwidth allocation", 
           "wrong flow type, this algorithm doesn't support 802.16j");
      }
    }
  }
  
  //
  //  downlink bandwidth allocation
  //

  
  // satisfy each dl flow class by class
  // ugs
  dl_free_symbol_num = dl_total_symbol_num;
  for( int i=0; i <  tmp_bs_to_ss.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_bs_to_ss.at(i);
    
    if( tmp_flow_queue->QoS_Class == 1 )
    {
      int allo_symbol_max = (int)floor((tmp_flow_queue->Rmax)
                                   /(8*PACKETS_PER_SECOND*tmp_flow_queue->rate));
      allo_symbol_max     = (int)ceil ( allo_symbol_max/ DL_PUSC_SLOT_SYMBOL_NUM) 
                                   * DL_PUSC_SLOT_SYMBOL_NUM;
                                   
      int allo_symbol = (int)ceil (tmp_flow_queue->queue_size()/tmp_flow_queue->rate);
      allo_symbol     = (int)ceil ( allo_symbol/ DL_PUSC_SLOT_SYMBOL_NUM) 
                                    * DL_PUSC_SLOT_SYMBOL_NUM;
      
      if(allo_symbol_max <= allo_symbol)
      {
        allo_symbol = allo_symbol_max;
      }
      if( dl_free_symbol_num <= allo_symbol)
      {
        allo_symbol = dl_free_symbol_num;
      }

      tmp_flow_queue->BWA_Total = allo_symbol;
        
      dl_free_symbol_num -= allo_symbol;
        
      if(dl_free_symbol_num <0)
      {
        lwx_err_msg("Strict Priority in ugs", "dl_free_symbol_num <0");
      }
      
    }
  }
  dl_total_symbol_num = dl_free_symbol_num;

  // ert-vr
  dl_free_symbol_num = dl_total_symbol_num;
  for( int i=0; i <  tmp_bs_to_ss.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_bs_to_ss.at(i);
    
    if( tmp_flow_queue->QoS_Class == 2 )
    {
      int allo_symbol_max = (int)floor((tmp_flow_queue->Rmax)
                                   /(8*PACKETS_PER_SECOND*tmp_flow_queue->rate));
      allo_symbol_max     = (int)ceil ( allo_symbol_max/ DL_PUSC_SLOT_SYMBOL_NUM) 
                                   * DL_PUSC_SLOT_SYMBOL_NUM;
                                   
      int allo_symbol = (int)ceil (tmp_flow_queue->queue_size()/tmp_flow_queue->rate);
      allo_symbol     = (int)ceil ( allo_symbol/ DL_PUSC_SLOT_SYMBOL_NUM) 
                                    * DL_PUSC_SLOT_SYMBOL_NUM;
      
      if(allo_symbol_max <= allo_symbol)
      {
        allo_symbol = allo_symbol_max;
      }
      if( dl_free_symbol_num <= allo_symbol)
      {
        allo_symbol = dl_free_symbol_num;
      }

      tmp_flow_queue->BWA_Total = allo_symbol;
        
      dl_free_symbol_num -= allo_symbol;
        
      if(dl_free_symbol_num <0)
      {
        lwx_err_msg("Strict Priority in ugs", "dl_free_symbol_num <0");
      }
      
    }
  }
  dl_total_symbol_num = dl_free_symbol_num;

  // rt-vr
  dl_free_symbol_num = dl_total_symbol_num;
  for( int i=0; i <  tmp_bs_to_ss.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_bs_to_ss.at(i);
    
    if( tmp_flow_queue->QoS_Class == 3 )
    {
      int allo_symbol_max = (int)floor((tmp_flow_queue->Rmax)
                                   /(8*PACKETS_PER_SECOND*tmp_flow_queue->rate));
      allo_symbol_max     = (int)ceil ( allo_symbol_max/ DL_PUSC_SLOT_SYMBOL_NUM) 
                                   * DL_PUSC_SLOT_SYMBOL_NUM;
                                   
      int allo_symbol = (int)ceil (tmp_flow_queue->queue_size()/tmp_flow_queue->rate);
      allo_symbol     = (int)ceil ( allo_symbol/ DL_PUSC_SLOT_SYMBOL_NUM) 
                                    * DL_PUSC_SLOT_SYMBOL_NUM;
      
      if(allo_symbol_max <= allo_symbol)
      {
        allo_symbol = allo_symbol_max;
      }
      if( dl_free_symbol_num <= allo_symbol)
      {
        allo_symbol = dl_free_symbol_num;
      }

      tmp_flow_queue->BWA_Total = allo_symbol;
        
      dl_free_symbol_num -= allo_symbol;
        
      if(dl_free_symbol_num <0)
      {
        lwx_err_msg("Strict Priority in ugs", "dl_free_symbol_num <0");
      }
      
    }
  }
  dl_total_symbol_num = dl_free_symbol_num;

  // nrt-vr
  dl_free_symbol_num = dl_total_symbol_num;
  for( int i=0; i <  tmp_bs_to_ss.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_bs_to_ss.at(i);
    
    if( tmp_flow_queue->QoS_Class == 4 )
    {
      int allo_symbol_max = (int)floor((tmp_flow_queue->Rmax)
                                   /(8*PACKETS_PER_SECOND*tmp_flow_queue->rate));
      allo_symbol_max     = (int)ceil ( allo_symbol_max/ DL_PUSC_SLOT_SYMBOL_NUM) 
                                   * DL_PUSC_SLOT_SYMBOL_NUM;
                                   
      int allo_symbol = (int)ceil (tmp_flow_queue->queue_size()/tmp_flow_queue->rate);
      allo_symbol     = (int)ceil ( allo_symbol/ DL_PUSC_SLOT_SYMBOL_NUM) 
                                    * DL_PUSC_SLOT_SYMBOL_NUM;
      
      if(allo_symbol_max <= allo_symbol)
      {
        allo_symbol = allo_symbol_max;
      }
      if( dl_free_symbol_num <= allo_symbol)
      {
        allo_symbol = dl_free_symbol_num;
      }

      tmp_flow_queue->BWA_Total = allo_symbol;
        
      dl_free_symbol_num -= allo_symbol;
        
      if(dl_free_symbol_num <0)
      {
        lwx_err_msg("Strict Priority in ugs", "dl_free_symbol_num <0");
      }
      
    }
  }
  dl_total_symbol_num = dl_free_symbol_num;

  // be
  dl_free_symbol_num = dl_total_symbol_num;
  for( int i=0; i <  tmp_bs_to_ss.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_bs_to_ss.at(i);
    
    if( tmp_flow_queue->QoS_Class == 5 )
    {
      int allo_symbol_max = (int)floor((tmp_flow_queue->Rmax)
                                   /(8*PACKETS_PER_SECOND*tmp_flow_queue->rate));
      allo_symbol_max     = (int)ceil ( allo_symbol_max/ DL_PUSC_SLOT_SYMBOL_NUM) 
                                   * DL_PUSC_SLOT_SYMBOL_NUM;
                                   
      int allo_symbol = (int)ceil (tmp_flow_queue->queue_size()/tmp_flow_queue->rate);
      allo_symbol     = (int)ceil ( allo_symbol/ DL_PUSC_SLOT_SYMBOL_NUM) 
                                    * DL_PUSC_SLOT_SYMBOL_NUM;
      
      if(allo_symbol_max <= allo_symbol)
      {
        allo_symbol = allo_symbol_max;
      }
      if( dl_free_symbol_num <= allo_symbol)
      {
        allo_symbol = dl_free_symbol_num;
      }

      tmp_flow_queue->BWA_Total = allo_symbol;
        
      dl_free_symbol_num -= allo_symbol;
        
      if(dl_free_symbol_num <0)
      {
        lwx_err_msg("Strict Priority in ugs", "dl_free_symbol_num <0");
      }
      
    }
  }
  dl_total_symbol_num = dl_free_symbol_num;

  // final dl BWA setting
  for( int i=0; i <  tmp_bs_to_ss.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_bs_to_ss.at(i);
    
    
    tmp_flow_queue->BWA_Total_Byte = (int)(tmp_flow_queue->BWA_Total 
                                     * tmp_flow_queue->rate + tmp_flow_queue->BWA_Backoff) ;
    tmp_flow_queue->BWA_Backoff = 0 ;
    tmp_flow_queue->BWA_Flow_Type = FLOW_TYPE_BS_TO_SS ;
    
  }
  
  
  // log each phase's bandwidth allocation of each downlink flow
  ostringstream dl_bwa_log;
  dl_bwa_log.clear();
  for( int i=0; i <  tmp_bs_to_ss.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_bs_to_ss.at(i);
    
    dl_bwa_log << tmp_flow_queue->src_ip << "/" 
               << tmp_flow_queue->src_port << "/" 
               << tmp_flow_queue->dst_ip << "/"  
               << tmp_flow_queue->dst_port << " " << endl;
    
    dl_bwa_log << "queue_size: " << tmp_flow_queue->queue_size() << endl;
    
    for(int j=0; j < tmp_flow_queue->BWA.size(); j++)
    { 
      dl_bwa_log << tmp_flow_queue->BWA.at(j) <<  "/" ;
    }
    dl_bwa_log << tmp_flow_queue->BWA_Total << "\n" ;
  }
  //lwx_test_msg("RR's DL BWA\n", dl_bwa_log.str());

  //
  //  uplink bandwidth allocation
  //

  // satisfy each ul flow class by class
  // ugs
  ul_free_symbol_num = ul_total_symbol_num;
  for( int i=0; i <  tmp_ss_to_bs.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ss_to_bs.at(i);
    
    if( tmp_flow_queue->QoS_Class == 1 )
    {
      int allo_symbol_max = (int)floor((tmp_flow_queue->Rmax)
                                   /(8*PACKETS_PER_SECOND*tmp_flow_queue->rate));
      allo_symbol_max     = (int)ceil ( allo_symbol_max/ UL_PUSC_SLOT_SYMBOL_NUM) 
                                   * UL_PUSC_SLOT_SYMBOL_NUM;
                                   
      int allo_symbol = (int)ceil (tmp_flow_queue->queue_size()/tmp_flow_queue->rate);
      allo_symbol     = (int)ceil ( allo_symbol/ UL_PUSC_SLOT_SYMBOL_NUM) 
                                    * UL_PUSC_SLOT_SYMBOL_NUM;
      
      if(allo_symbol_max <= allo_symbol)
      {
        allo_symbol = allo_symbol_max;
      }
      if( ul_free_symbol_num <= allo_symbol)
      {
        allo_symbol = ul_free_symbol_num;
      }

      tmp_flow_queue->BWA_Total = allo_symbol;
        
      ul_free_symbol_num -= allo_symbol;
        
      if(ul_free_symbol_num <0)
      {
        lwx_err_msg("Strict Priority in ugs", "ul_free_symbol_num <0");
      }
      
    }
  }
  ul_total_symbol_num = ul_free_symbol_num;

  // ert-vr
  ul_free_symbol_num = ul_total_symbol_num;
  for( int i=0; i <  tmp_ss_to_bs.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ss_to_bs.at(i);
    
    if( tmp_flow_queue->QoS_Class == 2 )
    {
      int allo_symbol_max = (int)floor((tmp_flow_queue->Rmax)
                                   /(8*PACKETS_PER_SECOND*tmp_flow_queue->rate));
      allo_symbol_max     = (int)ceil ( allo_symbol_max/ UL_PUSC_SLOT_SYMBOL_NUM) 
                                   * UL_PUSC_SLOT_SYMBOL_NUM;
                                   
      int allo_symbol = (int)ceil (tmp_flow_queue->queue_size()/tmp_flow_queue->rate);
      allo_symbol     = (int)ceil ( allo_symbol/ UL_PUSC_SLOT_SYMBOL_NUM) 
                                    * UL_PUSC_SLOT_SYMBOL_NUM;
      
      if(allo_symbol_max <= allo_symbol)
      {
        allo_symbol = allo_symbol_max;
      }
      if( ul_free_symbol_num <= allo_symbol)
      {
        allo_symbol = ul_free_symbol_num;
      }

      tmp_flow_queue->BWA_Total = allo_symbol;
        
      ul_free_symbol_num -= allo_symbol;
        
      if(ul_free_symbol_num <0)
      {
        lwx_err_msg("Strict Priority in ugs", "ul_free_symbol_num <0");
      }
      
    }
  }
  ul_total_symbol_num = ul_free_symbol_num;

  // rt-vr
  ul_free_symbol_num = ul_total_symbol_num;
  for( int i=0; i <  tmp_ss_to_bs.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ss_to_bs.at(i);
    
    if( tmp_flow_queue->QoS_Class == 3 )
    {
      int allo_symbol_max = (int)floor((tmp_flow_queue->Rmax)
                                   /(8*PACKETS_PER_SECOND*tmp_flow_queue->rate));
      allo_symbol_max     = (int)ceil ( allo_symbol_max/ UL_PUSC_SLOT_SYMBOL_NUM) 
                                   * UL_PUSC_SLOT_SYMBOL_NUM;
                                   
      int allo_symbol = (int)ceil (tmp_flow_queue->queue_size()/tmp_flow_queue->rate);
      allo_symbol     = (int)ceil ( allo_symbol/ UL_PUSC_SLOT_SYMBOL_NUM) 
                                    * UL_PUSC_SLOT_SYMBOL_NUM;
      
      if(allo_symbol_max <= allo_symbol)
      {
        allo_symbol = allo_symbol_max;
      }
      if( ul_free_symbol_num <= allo_symbol)
      {
        allo_symbol = ul_free_symbol_num;
      }

      tmp_flow_queue->BWA_Total = allo_symbol;
        
      ul_free_symbol_num -= allo_symbol;
        
      if(ul_free_symbol_num <0)
      {
        lwx_err_msg("Strict Priority in ugs", "ul_free_symbol_num <0");
      }
      
    }
  }
  ul_total_symbol_num = ul_free_symbol_num;

  // nrt-vr
  ul_free_symbol_num = ul_total_symbol_num;
  for( int i=0; i <  tmp_ss_to_bs.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ss_to_bs.at(i);
    
    if( tmp_flow_queue->QoS_Class == 4 )
    {
      int allo_symbol_max = (int)floor((tmp_flow_queue->Rmax)
                                   /(8*PACKETS_PER_SECOND*tmp_flow_queue->rate));
      allo_symbol_max     = (int)ceil ( allo_symbol_max/ UL_PUSC_SLOT_SYMBOL_NUM) 
                                   * UL_PUSC_SLOT_SYMBOL_NUM;
                                   
      int allo_symbol = (int)ceil (tmp_flow_queue->queue_size()/tmp_flow_queue->rate);
      allo_symbol     = (int)ceil ( allo_symbol/ UL_PUSC_SLOT_SYMBOL_NUM) 
                                    * UL_PUSC_SLOT_SYMBOL_NUM;
      
      if(allo_symbol_max <= allo_symbol)
      {
        allo_symbol = allo_symbol_max;
      }
      if( ul_free_symbol_num <= allo_symbol)
      {
        allo_symbol = ul_free_symbol_num;
      }

      tmp_flow_queue->BWA_Total = allo_symbol;
        
      ul_free_symbol_num -= allo_symbol;
        
      if(ul_free_symbol_num <0)
      {
        lwx_err_msg("Strict Priority in ugs", "ul_free_symbol_num <0");
      }
      
    }
  }
  ul_total_symbol_num = ul_free_symbol_num;

  // be
  ul_free_symbol_num = ul_total_symbol_num;
  for( int i=0; i <  tmp_ss_to_bs.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ss_to_bs.at(i);
    
    if( tmp_flow_queue->QoS_Class == 5 )
    {
      int allo_symbol_max = (int)floor((tmp_flow_queue->Rmax)
                                   /(8*PACKETS_PER_SECOND*tmp_flow_queue->rate));
      allo_symbol_max     = (int)ceil ( allo_symbol_max/ UL_PUSC_SLOT_SYMBOL_NUM) 
                                   * UL_PUSC_SLOT_SYMBOL_NUM;
                                   
      int allo_symbol = (int)ceil (tmp_flow_queue->queue_size()/tmp_flow_queue->rate);
      allo_symbol     = (int)ceil ( allo_symbol/ UL_PUSC_SLOT_SYMBOL_NUM) 
                                    * UL_PUSC_SLOT_SYMBOL_NUM;
      
      if(allo_symbol_max <= allo_symbol)
      {
        allo_symbol = allo_symbol_max;
      }
      if( ul_free_symbol_num <= allo_symbol)
      {
        allo_symbol = ul_free_symbol_num;
      }

      tmp_flow_queue->BWA_Total = allo_symbol;
        
      ul_free_symbol_num -= allo_symbol;
        
      if(ul_free_symbol_num <0)
      {
        lwx_err_msg("Strict Priority in ugs", "ul_free_symbol_num <0");
      }
    }
  }
  ul_total_symbol_num = ul_free_symbol_num;

  // final ul BWA setting
  for( int i=0; i <  tmp_ss_to_bs.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ss_to_bs.at(i);

    tmp_flow_queue->BWA_Total_Byte = (int)(tmp_flow_queue->BWA_Total 
                                 * tmp_flow_queue->rate + tmp_flow_queue->BWA_Backoff) ;
    tmp_flow_queue->BWA_Backoff = 0 ;
    tmp_flow_queue->BWA_Flow_Type=FLOW_TYPE_SS_TO_BS;
  }
  
  
  // log each phase's bandwidth allocation of each uplink flow
  ostringstream ul_bwa_log;
  ul_bwa_log.clear();
  for( int i=0; i <  tmp_ss_to_bs.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ss_to_bs.at(i);
    
    ul_bwa_log << tmp_flow_queue->src_ip << "/" 
               << tmp_flow_queue->src_port << "/" 
               << tmp_flow_queue->dst_ip << "/"  
               << tmp_flow_queue->dst_port << " " << endl;
  
    ul_bwa_log << "queue_size: " << tmp_flow_queue->queue_size() << endl;
               
    for(int j=0; j < tmp_flow_queue->BWA.size(); j++)
    { 
      ul_bwa_log << tmp_flow_queue->BWA.at(j) <<  "/" ;
    }
    
    ul_bwa_log << tmp_flow_queue->BWA_Total << "\n" ;
  }
  //lwx_test_msg("RR's UL BWA\n", ul_bwa_log.str());
  
}

// this algorithm only support transparent rs
// non-transparent rs is not supported
//
// algorithm:
//   relay path selection: 
//      relay link first
//   bandwidth allocation 
//      access: rr+qos min
//      relay:  rr

void LWX::BWA_Relay_Sample(int node_id)
{


  // create pointer to each flow type's queue to allocation bandwidth efficiencially
  vector <Flow_Queue*> tmp_dl_access_flow;
  vector <Flow_Queue*> tmp_dl_relay_flow;
  vector <Flow_Queue*> tmp_ul_access_flow;
  vector <Flow_Queue*> tmp_ul_relay_flow;
  
  
  // get symbol num of DL and UL
  int dl_access_zone_bwa  = DL_ACCESS_ZONE_SYMBOL * DL_PUSC_SUBCH_NUM ;
  int dl_relay_zone_bwa   = DL_RELAY_ZONE_SYMBOL * DL_PUSC_SUBCH_NUM ;
  int ul_access_zone_bwa  = UL_ACCESS_ZONE_SYMBOL * UL_PUSC_SUBCH_NUM ;
  int ul_relay_zone_bwa   = UL_RELAY_ZONE_SYMBOL * UL_PUSC_SUBCH_NUM ;
  
  // algorithm calculation var
  int dl_access_zone_bwa_free=0;
  int dl_relay_zone_bwa_free=0;
  int ul_access_zone_bwa_free=0;
  int ul_relay_zone_bwa_free=0;

  
  // reset all relative var's value in flow's attribute
  for(int i=0; i < lwx_node_list.size(); i++)
  {
    LWX_Node *tmp_node = &lwx_node_list.at(i);
    for( int j=0; j <  tmp_node->ctrl_flow.size(); j++)
    { 
      
      Flow_Queue *tmp_flow_queue = &tmp_node->ctrl_flow.at(j);
      tmp_flow_queue->BWA.clear();
      tmp_flow_queue->BWA_Total=-1;
      tmp_flow_queue->BWA_Rate=-1;
      tmp_flow_queue->BWA_Total_Byte=-1;
      tmp_flow_queue->BWA_Flow_Type=-1;
      tmp_flow_queue->BWA_Next_Hop=-1;
    }
  }



  // =========================================================
  // path selection:
  //   1.find the path (directly to bs or passing through rs)
  //   2.set BWA_Flow_Type and BWA_Next_Hop
  // =========================================================
  
  // set relay route's access link (second hop) to its own flow_type
  // because this algorithm only for two hop relay and the seconde hop
  // is defaultly set to ss or bs and flow_type is the default flow type
  // e.g. rs->bs or rs->ss_tp
  for(int i=0; i < lwx_node_list.size(); i++)
  {
    LWX_Node *tmp_node = &lwx_node_list.at(i);
    for( int j=0; j <  tmp_node->ctrl_flow.size(); j++)
    { 
      Flow_Queue *tmp_flow_queue = &tmp_node->ctrl_flow.at(j);
      switch(tmp_flow_queue->flow_type)
      {
        case FLOW_TYPE_RS_TO_BS:
          tmp_flow_queue->BWA_Flow_Type = tmp_flow_queue->flow_type;
          tmp_flow_queue->BWA_Next_Hop  = tmp_flow_queue->to_node_id;
          tmp_flow_queue->BWA_Rate      = tmp_flow_queue->rate;
          break;
        case FLOW_TYPE_RS_TO_SS_TP:
          tmp_flow_queue->BWA_Flow_Type = tmp_flow_queue->flow_type;
          tmp_flow_queue->BWA_Next_Hop  = tmp_flow_queue->to_node_id;
          tmp_flow_queue->BWA_Rate      = tmp_flow_queue->rate;
          break;
        default:
          break;
      }
      
    }
  }
  
  // path selection:
  //   # if this flow has relay link, use the first relay link as its next hop
  //   # else use access link
  for(int i=0; i < lwx_node_list.size(); i++)
  {
    LWX_Node *tmp_node = &lwx_node_list.at(i);
    for( int j=0; j <  tmp_node->ctrl_flow.size(); j++)
    { 
      Flow_Queue *tmp_flow_queue = &tmp_node->ctrl_flow.at(j);
      if(tmp_flow_queue->flow_type == FLOW_TYPE_BS_TO_SS)
      {
        if(!tmp_flow_queue->relay_link.empty())
        {
          tmp_flow_queue->BWA_Flow_Type = tmp_flow_queue->relay_link.at(0).flow_type;
          tmp_flow_queue->BWA_Next_Hop  = tmp_flow_queue->relay_link.at(0).rs_node_id;
          tmp_flow_queue->BWA_Rate      = tmp_flow_queue->relay_link.at(0).rate;
        }
        else
        {
          tmp_flow_queue->BWA_Flow_Type = tmp_flow_queue->flow_type;
          tmp_flow_queue->BWA_Next_Hop  = tmp_flow_queue->to_node_id;
          tmp_flow_queue->BWA_Rate      = tmp_flow_queue->rate;
        }
      }
      
      if(tmp_flow_queue->flow_type == FLOW_TYPE_SS_TO_BS)
      {
        if(!tmp_flow_queue->relay_link.empty())
        {
          tmp_flow_queue->BWA_Flow_Type = tmp_flow_queue->relay_link.at(0).flow_type;
          tmp_flow_queue->BWA_Next_Hop  = tmp_flow_queue->relay_link.at(0).rs_node_id;
          tmp_flow_queue->BWA_Rate      = tmp_flow_queue->relay_link.at(0).rate;
        }
        else
        {
          tmp_flow_queue->BWA_Flow_Type = tmp_flow_queue->flow_type;
          tmp_flow_queue->BWA_Next_Hop  = tmp_flow_queue->to_node_id;
          tmp_flow_queue->BWA_Rate      = tmp_flow_queue->rate;
        }
      }
      
    }
  }
  
  // =========================================================
  // bandwidth allocation:
  //   # access zone: RR + min QoS guarantee
  //   # relay zone:  only RR
  // =========================================================

  // collect flow into corresponding flow queue
  
  // !!!!!!! very important design: !!!!!!!!!!!
  //   bs->ss and bs->rs should be scheduling together, because they use same bandwidth
  //   ss->bs_tp and ss->rs should be also scheduling together, because they use same bandwidth
  
  for(int i=0; i < lwx_node_list.size(); i++)
  {
    LWX_Node *tmp_node = &lwx_node_list.at(i);
    for( int j=0; j <  tmp_node->ctrl_flow.size(); j++)
    { 
      switch(tmp_node->ctrl_flow.at(j).BWA_Flow_Type)
      {
        case FLOW_TYPE_BS_TO_SS:
        case FLOW_TYPE_RS_TO_SS_TP:
          tmp_dl_access_flow.push_back(&tmp_node->ctrl_flow.at(j));
          break;
        case FLOW_TYPE_SS_TO_BS:
        case FLOW_TYPE_RS_TO_BS:          
          tmp_ul_access_flow.push_back(&tmp_node->ctrl_flow.at(j));
          break;
        case FLOW_TYPE_SS_TO_RS_TP:
          tmp_ul_relay_flow.push_back(&tmp_node->ctrl_flow.at(j));
          break;
        case FLOW_TYPE_BS_TO_RS:
          tmp_dl_relay_flow.push_back(&tmp_node->ctrl_flow.at(j));
          break; 
        default:
          lwx_err_msg("in bandwidth allocation", 
             "wrong flow type, this algorithm doesn't support 802.16j non-transparent RS");
      }
      
    }
  }
  

  //
  //  downlink access bandwidth allocation
  //


  // satisfy each dl flow's Rmin
  dl_access_zone_bwa_free = dl_access_zone_bwa;
  for( int i=0; i <  tmp_dl_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_dl_access_flow.at(i);
    
    if( tmp_flow_queue->QoS_Class == 1 || tmp_flow_queue->QoS_Class == 2 ||
        tmp_flow_queue->QoS_Class == 3 || tmp_flow_queue->QoS_Class == 4 )
    {
      int allo_symbol = (int)ceil((tmp_flow_queue->Rmin)
                              /(8*PACKETS_PER_SECOND*tmp_flow_queue->BWA_Rate));
      allo_symbol     = (int)ceil ( allo_symbol/ DL_PUSC_SLOT_SYMBOL_NUM) 
                                * DL_PUSC_SLOT_SYMBOL_NUM;
      
      if(allo_symbol > dl_access_zone_bwa_free)
      {
        allo_symbol=dl_access_zone_bwa_free; 
      }
      
      tmp_flow_queue->BWA.push_back(allo_symbol);
      tmp_flow_queue->BWA_Total+=allo_symbol;
      
      dl_access_zone_bwa_free -= allo_symbol;
      
      if(dl_access_zone_bwa_free <0)
      {
        lwx_err_msg("RR's DL Rmin", "dl_free_symbol_num <0");
      }
      
    }
  }
  dl_access_zone_bwa = dl_access_zone_bwa_free;
  
  
  // satisfy each dl flow by RR
  dl_access_zone_bwa_free = dl_access_zone_bwa;
  for( int i=0; i <  tmp_dl_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_dl_access_flow.at(i);
    
    int allo_symbol_max = (int)floor((tmp_flow_queue->Rmax)
                                 /(8*PACKETS_PER_SECOND*tmp_flow_queue->BWA_Rate));
    allo_symbol_max     = (int)ceil ( allo_symbol_max/ DL_PUSC_SLOT_SYMBOL_NUM) 
                                 * DL_PUSC_SLOT_SYMBOL_NUM;
    int allo_symbol = (int)ceil ( dl_access_zone_bwa_free / (tmp_dl_access_flow.size()-i));
    allo_symbol     = (int)ceil ( allo_symbol/ DL_PUSC_SLOT_SYMBOL_NUM) 
                             * DL_PUSC_SLOT_SYMBOL_NUM;
      
    if(allo_symbol_max <= allo_symbol + tmp_flow_queue->BWA_Total)
    {
      allo_symbol = allo_symbol_max - tmp_flow_queue->BWA_Total;
    }
    tmp_flow_queue->BWA_Total += allo_symbol;
      
    dl_access_zone_bwa_free -= allo_symbol;
      
    if(dl_access_zone_bwa_free <0)
    {
      lwx_err_msg("RR's DL RR", "dl_access_zone_bwa_free <0");
    }

  }
  dl_access_zone_bwa = dl_access_zone_bwa_free;
  

  // final dl BWA setting
  for( int i=0; i <  tmp_dl_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_dl_access_flow.at(i);
    
    
    tmp_flow_queue->BWA_Total_Byte = (int)(tmp_flow_queue->BWA_Total 
                                     * tmp_flow_queue->BWA_Rate+ tmp_flow_queue->BWA_Backoff) ;
    tmp_flow_queue->BWA_Backoff = 0 ;
    
  }
  
  
  // log each phase's bandwidth allocation of each downlink flow
  ostringstream dl_access_bwa_log;
  dl_access_bwa_log << "============ dl access bwa log: ============" << endl;
  dl_access_bwa_log << "{(src_node_id/from_node_id/to_node_id) (src_ip/src_port/dst_port)\n"
                    << " (flow_type/rate) qsize (queue_size) \n"
                    << "BWA: (Total/Total_Byte/Flow_type/Next_Hop/Rate)} \n";
  
  for( int i=0; i <  tmp_dl_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_dl_access_flow.at(i);
    
    dl_access_bwa_log << "{("
                      << tmp_flow_queue->src_node_id << "/" 
                      << tmp_flow_queue->from_node_id << "/" 
                      << tmp_flow_queue->to_node_id << "/"  
                      << tmp_flow_queue->dst_port << ") " ;
    
    dl_access_bwa_log << "(" << tmp_flow_queue->src_ip << "/" 
                      << tmp_flow_queue->src_port << "/" 
                      << tmp_flow_queue->dst_ip << "/"  
                      << tmp_flow_queue->dst_port << ") ";
                      
    dl_access_bwa_log << "(" << tmp_flow_queue->flow_type << "/" 
                      << tmp_flow_queue->rate << ") ";
    
    dl_access_bwa_log << "qsize (" << tmp_flow_queue->queue_size() << ") " ;
    
    dl_access_bwa_log << "BWA: (" 
                      << tmp_flow_queue->BWA_Total << "/" 
                      << tmp_flow_queue->BWA_Total_Byte << "/" 
                      << tmp_flow_queue->BWA_Flow_Type << "/" 
                      << tmp_flow_queue->BWA_Next_Hop << "/" 
                      << tmp_flow_queue->BWA_Rate << ")} " ;
                      
  }
  //lwx_test_msg("RR's DL ACCESS BWA\n", dl_access_bwa_log.str());

  //
  //  downlink relay bandwidth allocation (rs->ss)
  //

  
  // satisfy each dl relay flow by RR
  dl_relay_zone_bwa_free = dl_relay_zone_bwa;
  for( int i=0; i <  tmp_dl_relay_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_dl_relay_flow.at(i);
    
    int allo_symbol_max = (int)floor((tmp_flow_queue->Rmax)
                                 /(8*PACKETS_PER_SECOND*tmp_flow_queue->BWA_Rate));
    allo_symbol_max     = (int)ceil ( allo_symbol_max/ DL_PUSC_SLOT_SYMBOL_NUM) 
                                 * DL_PUSC_SLOT_SYMBOL_NUM;
    int allo_symbol = (int)ceil ( dl_relay_zone_bwa_free / (tmp_dl_relay_flow.size()-i));
    allo_symbol     = (int)ceil ( allo_symbol/ DL_PUSC_SLOT_SYMBOL_NUM) 
                             * DL_PUSC_SLOT_SYMBOL_NUM;
      
    if(allo_symbol_max <= allo_symbol + tmp_flow_queue->BWA_Total)
    {
      allo_symbol = allo_symbol_max - tmp_flow_queue->BWA_Total;
    }
    tmp_flow_queue->BWA_Total += allo_symbol;
      
    dl_relay_zone_bwa_free -= allo_symbol;
      
    if(dl_relay_zone_bwa_free <0)
    {
      lwx_err_msg("RR's DL Relay RR", "dl_relay_zone_bwa_free <0");
    }

  }
  dl_relay_zone_bwa = dl_relay_zone_bwa_free;
  

  // final dl BWA setting
  for( int i=0; i <  tmp_dl_relay_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_dl_relay_flow.at(i);
    
    
    tmp_flow_queue->BWA_Total_Byte = (int)(tmp_flow_queue->BWA_Total 
                                     * tmp_flow_queue->BWA_Rate + tmp_flow_queue->BWA_Backoff) ;
    tmp_flow_queue->BWA_Backoff = 0 ;
    
  }
  
  // log each phase's bandwidth allocation of each downlink relay flow
  ostringstream dl_relay_bwa_log;
  dl_relay_bwa_log << "============ dl relay bwa log: ============" << endl;
  dl_relay_bwa_log << "{(src_node_id/from_node_id/to_node_id) (src_ip/src_port/dst_port)\n"
                    << " (flow_type/rate) qsize (queue_size) \n"
                    << "BWA: (Total/Total_Byte/Flow_type/Next_Hop/Rate)} \n";
  
  for( int i=0; i <  tmp_dl_relay_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_dl_relay_flow.at(i);
    
    dl_relay_bwa_log  << "{("
                      << tmp_flow_queue->src_node_id << "/" 
                      << tmp_flow_queue->from_node_id << "/" 
                      << tmp_flow_queue->to_node_id << "/"  
                      << tmp_flow_queue->dst_port << ") " ;
    
    dl_relay_bwa_log  << "(" << tmp_flow_queue->src_ip << "/" 
                      << tmp_flow_queue->src_port << "/" 
                      << tmp_flow_queue->dst_ip << "/"  
                      << tmp_flow_queue->dst_port << ") ";
                      
    dl_relay_bwa_log  << "(" << tmp_flow_queue->flow_type << "/" 
                      << tmp_flow_queue->rate << ") ";
    
    dl_relay_bwa_log << "qsize (" << tmp_flow_queue->queue_size() << ") " ;
    
    dl_relay_bwa_log << "BWA: (" 
                      << tmp_flow_queue->BWA_Total << "/" 
                      << tmp_flow_queue->BWA_Total_Byte << "/" 
                      << tmp_flow_queue->BWA_Flow_Type << "/" 
                      << tmp_flow_queue->BWA_Next_Hop << "/" 
                      << tmp_flow_queue->BWA_Rate << ")} " ;
                      
  }
  //lwx_test_msg("RR's DL RELAY BWA\n", dl_relay_bwa_log.str());


  //
  //  uplink access bandwidth allocation
  //


  // satisfy each ul flow's Rmin
  ul_access_zone_bwa_free = ul_access_zone_bwa;
  for( int i=0; i <  tmp_ul_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ul_access_flow.at(i);
    
    if( tmp_flow_queue->QoS_Class == 1 || tmp_flow_queue->QoS_Class == 2 ||
        tmp_flow_queue->QoS_Class == 3 || tmp_flow_queue->QoS_Class == 4 )
    {
      int allo_symbol = (int)ceil((tmp_flow_queue->Rmin)
                              /(8*PACKETS_PER_SECOND*tmp_flow_queue->BWA_Rate));
      allo_symbol     = (int)ceil ( allo_symbol/ UL_PUSC_SLOT_SYMBOL_NUM) 
                                * UL_PUSC_SLOT_SYMBOL_NUM;
      
      if(allo_symbol > ul_access_zone_bwa_free)
      {
        allo_symbol=ul_access_zone_bwa_free; 
      }
      
      tmp_flow_queue->BWA.push_back(allo_symbol);
      tmp_flow_queue->BWA_Total+=allo_symbol;
      
      ul_access_zone_bwa_free -= allo_symbol;
      
      if(ul_access_zone_bwa_free <0)
      {
        lwx_err_msg("RR's UL Rmin", "ul_free_symbol_num <0");
      }
      
    }
  }
  ul_access_zone_bwa = ul_access_zone_bwa_free;
  
  
  // satisfy each ul flow by RR
  ul_access_zone_bwa_free = ul_access_zone_bwa;
  for( int i=0; i <  tmp_ul_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ul_access_flow.at(i);
    
    int allo_symbol_max = (int)floor((tmp_flow_queue->Rmax)
                                 /(8*PACKETS_PER_SECOND*tmp_flow_queue->BWA_Rate));
    allo_symbol_max     = (int)ceil ( allo_symbol_max/ UL_PUSC_SLOT_SYMBOL_NUM) 
                                 * UL_PUSC_SLOT_SYMBOL_NUM;
    int allo_symbol = (int)ceil ( ul_access_zone_bwa_free / (tmp_ul_access_flow.size()-i));
    allo_symbol     = (int)ceil ( allo_symbol/ UL_PUSC_SLOT_SYMBOL_NUM) 
                             * UL_PUSC_SLOT_SYMBOL_NUM;
      
    if(allo_symbol_max <= allo_symbol + tmp_flow_queue->BWA_Total)
    {
      allo_symbol = allo_symbol_max - tmp_flow_queue->BWA_Total;
    }
    tmp_flow_queue->BWA_Total += allo_symbol;
      
    ul_access_zone_bwa_free -= allo_symbol;
      
    if(ul_access_zone_bwa_free <0)
    {
      lwx_err_msg("RR's UL RR", "ul_access_zone_bwa_free <0");
    }

  }
  ul_access_zone_bwa = ul_access_zone_bwa_free;
  

  // final ul BWA setting
  for( int i=0; i <  tmp_ul_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ul_access_flow.at(i);
    
    
    tmp_flow_queue->BWA_Total_Byte = (int)(tmp_flow_queue->BWA_Total 
                                     * tmp_flow_queue->BWA_Rate + tmp_flow_queue->BWA_Backoff) ;
    tmp_flow_queue->BWA_Backoff = 0 ;
    
  }
  
  
  // log each phase's bandwidth allocation of each downlink flow
  ostringstream ul_access_bwa_log;
  ul_access_bwa_log << "============ ul access bwa log: ============" << endl;
  ul_access_bwa_log << "{(src_node_id/from_node_id/to_node_id) (src_ip/src_port/dst_port)\n"
                    << " (flow_type/rate) qsize (queue_size) \n"
                    << "BWA: (Total/Total_Byte/Flow_type/Next_Hop/Rate)} \n";
  
  for( int i=0; i <  tmp_ul_access_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ul_access_flow.at(i);
    
    ul_access_bwa_log << "{("
                      << tmp_flow_queue->src_node_id << "/" 
                      << tmp_flow_queue->from_node_id << "/" 
                      << tmp_flow_queue->to_node_id << "/"  
                      << tmp_flow_queue->dst_port << ") " ;
    
    ul_access_bwa_log << "(" << tmp_flow_queue->src_ip << "/" 
                      << tmp_flow_queue->src_port << "/" 
                      << tmp_flow_queue->dst_ip << "/"  
                      << tmp_flow_queue->dst_port << ") ";
                      
    ul_access_bwa_log << "(" << tmp_flow_queue->flow_type << "/" 
                      << tmp_flow_queue->rate << ") ";
    
    ul_access_bwa_log << "qsize (" << tmp_flow_queue->queue_size() << ") " ;
    
    ul_access_bwa_log << "BWA: (" 
                      << tmp_flow_queue->BWA_Total << "/" 
                      << tmp_flow_queue->BWA_Total_Byte << "/" 
                      << tmp_flow_queue->BWA_Flow_Type << "/" 
                      << tmp_flow_queue->BWA_Next_Hop << "/" 
                      << tmp_flow_queue->BWA_Rate << ")} " ;
                      
  }
  //lwx_test_msg("RR's UL ACCESS BWA\n", ul_access_bwa_log.str());

  //
  //  uplink relay bandwidth allocation (rs->ss)
  //

  
  // satisfy each ul relay flow by RR
  ul_relay_zone_bwa_free = ul_relay_zone_bwa;
  for( int i=0; i <  tmp_ul_relay_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ul_relay_flow.at(i);
    
    int allo_symbol_max = (int)floor((tmp_flow_queue->Rmax)
                                 /(8*PACKETS_PER_SECOND*tmp_flow_queue->BWA_Rate));
    allo_symbol_max     = (int)ceil ( allo_symbol_max/ UL_PUSC_SLOT_SYMBOL_NUM) 
                                 * UL_PUSC_SLOT_SYMBOL_NUM;
    int allo_symbol = (int)ceil ( ul_relay_zone_bwa_free / (tmp_ul_relay_flow.size()-i));
    allo_symbol     = (int)ceil ( allo_symbol/ UL_PUSC_SLOT_SYMBOL_NUM) 
                             * UL_PUSC_SLOT_SYMBOL_NUM;
      
    if(allo_symbol_max <= allo_symbol + tmp_flow_queue->BWA_Total)
    {
      allo_symbol = allo_symbol_max - tmp_flow_queue->BWA_Total;
    }
    tmp_flow_queue->BWA_Total += allo_symbol;
      
    ul_relay_zone_bwa_free -= allo_symbol;
      
    if(ul_relay_zone_bwa_free <0)
    {
      lwx_err_msg("RR's UL Relay RR", "ul_relay_zone_bwa_free <0");
    }

  }
  ul_relay_zone_bwa = ul_relay_zone_bwa_free;
  

  // final ul BWA setting
  for( int i=0; i <  tmp_ul_relay_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ul_relay_flow.at(i);
    
    
    tmp_flow_queue->BWA_Total_Byte = (int)(tmp_flow_queue->BWA_Total 
                                     * tmp_flow_queue->BWA_Rate + tmp_flow_queue->BWA_Backoff) ;
    tmp_flow_queue->BWA_Backoff = 0 ;
    
  }
  
  // log each phase's bandwidth allocation of each downlink relay flow
  ostringstream ul_relay_bwa_log;
  ul_relay_bwa_log << "============ ul relay bwa log: ============" << endl;
  ul_relay_bwa_log << "{(src_node_id/from_node_id/to_node_id) (src_ip/src_port/dst_port)\n"
                    << " (flow_type/rate) qsize (queue_size) \n"
                    << "BWA: (Total/Total_Byte/Flow_type/Next_Hop/Rate)} \n";
  
  for( int i=0; i <  tmp_ul_relay_flow.size(); i++)
  { 
    Flow_Queue *tmp_flow_queue = tmp_ul_relay_flow.at(i);
    
    ul_relay_bwa_log  << "{("
                      << tmp_flow_queue->src_node_id << "/" 
                      << tmp_flow_queue->from_node_id << "/" 
                      << tmp_flow_queue->to_node_id << "/"  
                      << tmp_flow_queue->dst_port << ") " ;
    
    ul_relay_bwa_log  << "(" << tmp_flow_queue->src_ip << "/" 
                      << tmp_flow_queue->src_port << "/" 
                      << tmp_flow_queue->dst_ip << "/"  
                      << tmp_flow_queue->dst_port << ") ";
                      
    ul_relay_bwa_log  << "(" << tmp_flow_queue->flow_type << "/" 
                      << tmp_flow_queue->rate << ") ";
    
    ul_relay_bwa_log << "qsize (" << tmp_flow_queue->queue_size() << ") " ;
    
    ul_relay_bwa_log << "BWA: (" 
                      << tmp_flow_queue->BWA_Total << "/" 
                      << tmp_flow_queue->BWA_Total_Byte << "/" 
                      << tmp_flow_queue->BWA_Flow_Type << "/" 
                      << tmp_flow_queue->BWA_Next_Hop << "/" 
                      << tmp_flow_queue->BWA_Rate << ")} " ;
                      
  }
  //lwx_test_msg("RR's UL RELAY BWA\n", ul_relay_bwa_log.str());


} 


/////////////////////
//  Packet Transmssion 
/////////////////////

void LWX::trans_bs_to_ss(int node_id, vector <Flow_Queue*> &tmp_flow_list)
{
  
  for( int i=0; i <  tmp_flow_list.size(); i++)
  {     
    
    Flow_Queue *tmp_flow_queue = tmp_flow_list.at(i);
    
    if(tmp_flow_queue->from_node_id == node_id)
    {
      // log test msg

      //ostringstream log_send_pkt ;
      //log_send_pkt << tmp_flow_queue->queue_length();

  
      while(tmp_flow_queue->queue_length()!=0)
      {
        // get packet
        Packet *tmp_pkt=tmp_flow_queue->pkt_front();
  
        // get log
        
        ostringstream log_bwa;
        log_bwa << tmp_flow_queue->src_ip << "/" << tmp_flow_queue->src_port << "/"
                << tmp_flow_queue->dst_ip << "/" << tmp_flow_queue->dst_port << " "
                << "BWA_Byte " << tmp_flow_queue->BWA_Total_Byte << " "
                << "pkt_size " << hdr_cmn::access(tmp_pkt)->size();
        
        //lwx_test_msg("dl", log_bwa.str());
        
        // decrease BWA_Total
        if((tmp_flow_queue->BWA_Total_Byte - hdr_cmn::access(tmp_pkt)->size())<0 )
        {
          tmp_flow_queue->BWA_Backoff = tmp_flow_queue->BWA_Total_Byte; // haha XD
          break; 
        }
        tmp_flow_queue->BWA_Total_Byte -= hdr_cmn::access(tmp_pkt)->size();
        
     
        // pop this packet
        tmp_flow_queue->pkt_pop();
   if (Scheduler::instance().clock() - hdr_lwx::access(tmp_pkt)->sendtime_ > hdr_lwx::access(tmp_pkt)->maxl_ )
      { 
        //drop packet added by pandeli
       drop(tmp_pkt, "DROP_RTR_TTL");
	}
	else {
        // send down packet added by pandeli
        downtarget_->recv(tmp_pkt,(Handler*) 0);
        }
        
      }
    }
  }
  
}

void LWX::trans_ss_to_bs(int node_id, vector <Flow_Queue*> &tmp_flow_list)
{
  for( int i=0; i <  tmp_flow_list.size(); i++)
  {     
    
    Flow_Queue *tmp_flow_queue = tmp_flow_list.at(i);
    
    if(tmp_flow_queue->from_node_id == node_id)
    {
      // log test msg

      //ostringstream log_send_pkt ;
      //log_send_pkt << tmp_flow_queue->queue_length();

  
      while(tmp_flow_queue->queue_length()!=0)
      {
        // get packet
        Packet *tmp_pkt=tmp_flow_queue->pkt_front();
  
        // get log
        
        ostringstream log_bwa;
        log_bwa << tmp_flow_queue->src_ip << "/" << tmp_flow_queue->src_port << "/"
                << tmp_flow_queue->dst_ip << "/" << tmp_flow_queue->dst_port << " "
                << "BWA_Byte " << tmp_flow_queue->BWA_Total_Byte << " "
                << "pkt_size " << hdr_cmn::access(tmp_pkt)->size();
        
        //lwx_test_msg("dl", log_bwa.str());
        
        // decrease BWA_Total
        if((tmp_flow_queue->BWA_Total_Byte - hdr_cmn::access(tmp_pkt)->size())<0 )
        {
          tmp_flow_queue->BWA_Backoff = tmp_flow_queue->BWA_Total_Byte; // haha XD
          break; 
        }
        tmp_flow_queue->BWA_Total_Byte -= hdr_cmn::access(tmp_pkt)->size();
        
     
        // pop this packet
        tmp_flow_queue->pkt_pop();
        
   if (Scheduler::instance().clock() - hdr_lwx::access(tmp_pkt)->sendtime_ > hdr_lwx::access(tmp_pkt)->maxl_ )
      { 
        //drop packet added by pandeli
       drop(tmp_pkt, "DROP_RTR_TTL");
	}
	else {
        // send down packet
        downtarget_->recv(tmp_pkt,(Handler*) 0);
        }
      }
    }
  }
}

void LWX::trans_bs_to_rs(int node_id, vector <Flow_Queue*> &tmp_flow_list)
{
  
  for( int i=0; i <  tmp_flow_list.size(); i++)
  {     
    
    Flow_Queue *tmp_flow_queue = tmp_flow_list.at(i);
    
    if(tmp_flow_queue->from_node_id == node_id)
    {
      // log test msg

      while(tmp_flow_queue->queue_length()!=0)
      {
        // get packet
        Packet *tmp_pkt=tmp_flow_queue->pkt_front();

        // decrease BWA_Total
        if((tmp_flow_queue->BWA_Total_Byte - hdr_cmn::access(tmp_pkt)->size())<0 )
        {
          tmp_flow_queue->BWA_Backoff = tmp_flow_queue->BWA_Total_Byte; // haha XD
          break; 
        }
        tmp_flow_queue->BWA_Total_Byte -= hdr_cmn::access(tmp_pkt)->size();
     
        // pop this packet
        tmp_flow_queue->pkt_pop();
        
        // get this pkt's next hop's mac
        // chang pkt direction from up to down
        // and change this pkt's mac_col->macDA_ 
        // and change lwx from_node_id and bwa_flow_type
        hdr_cmn::access(tmp_pkt)->direction() = hdr_cmn::DOWN;
        hdr_mac::access(tmp_pkt)->macDA_ = find_mac_addr(tmp_flow_queue->BWA_Next_Hop);
        hdr_lwx::access(tmp_pkt)->from_node_id()  = tmp_flow_queue->from_node_id;
        hdr_lwx::access(tmp_pkt)->bwa_flow_type() = tmp_flow_queue->BWA_Flow_Type;
        
   if (Scheduler::instance().clock() - hdr_lwx::access(tmp_pkt)->sendtime_ > hdr_lwx::access(tmp_pkt)->maxl_ )
      { 
        //drop packet added by pandeli
       drop(tmp_pkt, "DROP_RTR_TTL");
	}
	else {
        // send down packet
        downtarget_->recv(tmp_pkt,(Handler*) 0);
        }
      }
    }
  }

}
  
void LWX::trans_rs_to_bs(int node_id, vector <Flow_Queue*> &tmp_flow_list)
{
  for( int i=0; i <  tmp_flow_list.size(); i++)
  {     
    
    Flow_Queue *tmp_flow_queue = tmp_flow_list.at(i);
    
    if(tmp_flow_queue->from_node_id == node_id)
    {
      // log test msg

      while(tmp_flow_queue->queue_length()!=0)
      {
        // get packet
        Packet *tmp_pkt=tmp_flow_queue->pkt_front();

        // decrease BWA_Total
        if((tmp_flow_queue->BWA_Total_Byte - hdr_cmn::access(tmp_pkt)->size())<0 )
        {
          tmp_flow_queue->BWA_Backoff = tmp_flow_queue->BWA_Total_Byte; // haha XD
          break; 
        }
        tmp_flow_queue->BWA_Total_Byte -= hdr_cmn::access(tmp_pkt)->size();
     
        // pop this packet added by pandeli
        tmp_flow_queue->pkt_pop();
        
        // get this pkt's next hop's mac
        // chang pkt direction from up to down
        // and change this pkt's mac_col->macDA_ 
        // and change lwx from_node_id and bwa_flow_type
        hdr_cmn::access(tmp_pkt)->direction() = hdr_cmn::DOWN;
        hdr_mac::access(tmp_pkt)->macDA_ = find_mac_addr(tmp_flow_queue->BWA_Next_Hop);
        hdr_lwx::access(tmp_pkt)->from_node_id()  = tmp_flow_queue->from_node_id;
        hdr_lwx::access(tmp_pkt)->bwa_flow_type() = tmp_flow_queue->BWA_Flow_Type;
        
   if (Scheduler::instance().clock() - hdr_lwx::access(tmp_pkt)->sendtime_ > hdr_lwx::access(tmp_pkt)->maxl_ )
      {   
        //drop packet added by pandeli
       drop(tmp_pkt, "DROP_RTR_TTL");
	}
	else {
        // send down packet
        downtarget_->recv(tmp_pkt,(Handler*) 0);
        }
      }
    }
  }
}

void LWX::trans_rs_to_ss_tp(int node_id, vector <Flow_Queue*> &tmp_flow_list)
{
  for( int i=0; i <  tmp_flow_list.size(); i++)
  {     
    
    Flow_Queue *tmp_flow_queue = tmp_flow_list.at(i);
    
    if(tmp_flow_queue->from_node_id == node_id)
    {
      // log test msg

      while(tmp_flow_queue->queue_length()!=0)
      {
        // get packet
        Packet *tmp_pkt=tmp_flow_queue->pkt_front();

        // decrease BWA_Total
        if((tmp_flow_queue->BWA_Total_Byte - hdr_cmn::access(tmp_pkt)->size())<0 )
        {
          tmp_flow_queue->BWA_Backoff = tmp_flow_queue->BWA_Total_Byte; // haha XD
          break; 
        }
        tmp_flow_queue->BWA_Total_Byte -= hdr_cmn::access(tmp_pkt)->size();
     
        // pop this packet
        tmp_flow_queue->pkt_pop();
        

        // get this pkt's next hop's mac
        // chang pkt direction from up to down
        // and change this pkt's mac_col->macDA_ 
        // and change lwx from_node_id and bwa_flow_type
        hdr_cmn::access(tmp_pkt)->direction() = hdr_cmn::DOWN;
        hdr_mac::access(tmp_pkt)->macDA_ = find_mac_addr(tmp_flow_queue->BWA_Next_Hop);
        hdr_lwx::access(tmp_pkt)->from_node_id()  = tmp_flow_queue->from_node_id;
        hdr_lwx::access(tmp_pkt)->bwa_flow_type() = tmp_flow_queue->BWA_Flow_Type;

   if (Scheduler::instance().clock() - hdr_lwx::access(tmp_pkt)->sendtime_ > hdr_lwx::access(tmp_pkt)->maxl_ )
      {    
        //drop packet - added by Pandeli
        drop(tmp_pkt, "DROP_RTR_TTL");
	}
	else {
        // send down packet
        downtarget_->recv(tmp_pkt,(Handler*) 0);
        }
        
      }
    }
  }
}
 
void LWX::trans_rs_to_ss_nontp(int node_id, vector <Flow_Queue*> &tmp_flow_list)
{
  for( int i=0; i <  tmp_flow_list.size(); i++)
  {     
    
    Flow_Queue *tmp_flow_queue = tmp_flow_list.at(i);
    
    if(tmp_flow_queue->from_node_id == node_id)
    {
      // log test msg

      while(tmp_flow_queue->queue_length()!=0)
      {
        // get packet
        Packet *tmp_pkt=tmp_flow_queue->pkt_front();

        // decrease BWA_Total
        if((tmp_flow_queue->BWA_Total_Byte - hdr_cmn::access(tmp_pkt)->size())<0 )
        {
          tmp_flow_queue->BWA_Backoff = tmp_flow_queue->BWA_Total_Byte; // haha XD
          break; 
        }
        tmp_flow_queue->BWA_Total_Byte -= hdr_cmn::access(tmp_pkt)->size();
     
        // pop this packet
        tmp_flow_queue->pkt_pop();
        
        // get this pkt's next hop's mac
        // chang pkt direction from up to down
        // and change this pkt's mac_col->macDA_ 
        // and change lwx from_node_id and bwa_flow_type
        hdr_cmn::access(tmp_pkt)->direction() = hdr_cmn::DOWN;
        hdr_mac::access(tmp_pkt)->macDA_ = find_mac_addr(tmp_flow_queue->BWA_Next_Hop);
        hdr_lwx::access(tmp_pkt)->from_node_id()  = tmp_flow_queue->from_node_id;
        hdr_lwx::access(tmp_pkt)->bwa_flow_type() = tmp_flow_queue->BWA_Flow_Type;
        
   if (Scheduler::instance().clock() - hdr_lwx::access(tmp_pkt)->sendtime_ > hdr_lwx::access(tmp_pkt)->maxl_ )
      { 
        //drop packet - added by pandeli
        drop(tmp_pkt, "DROP_RTR_TTL");
	}
	else {
        // send down packet
        downtarget_->recv(tmp_pkt,(Handler*) 0);
        }
        
      }
    }
  }
}

void LWX::trans_ss_to_rs_tp(int node_id, vector <Flow_Queue*> &tmp_flow_list)
{
  for( int i=0; i <  tmp_flow_list.size(); i++)
  {     
    
    Flow_Queue *tmp_flow_queue = tmp_flow_list.at(i);
    
    if(tmp_flow_queue->from_node_id == node_id)
    {
      // log test msg

      while(tmp_flow_queue->queue_length()!=0)
      {
        // get packet
        Packet *tmp_pkt=tmp_flow_queue->pkt_front();

        // decrease BWA_Total
        if((tmp_flow_queue->BWA_Total_Byte - hdr_cmn::access(tmp_pkt)->size())<0 )
        {
          tmp_flow_queue->BWA_Backoff = tmp_flow_queue->BWA_Total_Byte; // haha XD
          break; 
        }
        tmp_flow_queue->BWA_Total_Byte -= hdr_cmn::access(tmp_pkt)->size();
     
        // pop this packet
        tmp_flow_queue->pkt_pop();
        
        // get this pkt's next hop's mac
        // chang pkt direction from up to down
        // and change this pkt's mac_col->macDA_ 
        // and change lwx from_node_id and bwa_flow_type
        hdr_cmn::access(tmp_pkt)->direction() = hdr_cmn::DOWN;
        hdr_mac::access(tmp_pkt)->macDA_ = find_mac_addr(tmp_flow_queue->BWA_Next_Hop);
        hdr_lwx::access(tmp_pkt)->from_node_id()  = tmp_flow_queue->from_node_id;
        hdr_lwx::access(tmp_pkt)->bwa_flow_type() = tmp_flow_queue->BWA_Flow_Type;
        
   if (Scheduler::instance().clock() - hdr_lwx::access(tmp_pkt)->sendtime_ > hdr_lwx::access(tmp_pkt)->maxl_ )
      { 
        //drop packet aded by pandeli
        drop(tmp_pkt, "DROP_RTR_TTL");
	}
	else {
        // send down packet
        downtarget_->recv(tmp_pkt,(Handler*) 0);
        }
        
      }
    }
  }
  
}
 
void LWX::trans_ss_to_rs_nontp(int node_id, vector <Flow_Queue*> &tmp_flow_list)
{
  for( int i=0; i <  tmp_flow_list.size(); i++)
  {     
    
    Flow_Queue *tmp_flow_queue = tmp_flow_list.at(i);
    
    if(tmp_flow_queue->from_node_id == node_id)
    {
      // log test msg

      while(tmp_flow_queue->queue_length()!=0)
      {
        // get packet
        Packet *tmp_pkt=tmp_flow_queue->pkt_front();

        // decrease BWA_Total
        if((tmp_flow_queue->BWA_Total_Byte - hdr_cmn::access(tmp_pkt)->size())<0 )
        {
          tmp_flow_queue->BWA_Backoff = tmp_flow_queue->BWA_Total_Byte; // haha XD
          break; 
        }
        tmp_flow_queue->BWA_Total_Byte -= hdr_cmn::access(tmp_pkt)->size();
     
        // pop this packet
        tmp_flow_queue->pkt_pop();
        
        // get this pkt's next hop's mac
        // chang pkt direction from up to down
        // and change this pkt's mac_col->macDA_ 
        // and change lwx from_node_id and bwa_flow_type
        hdr_cmn::access(tmp_pkt)->direction() = hdr_cmn::DOWN;
        hdr_mac::access(tmp_pkt)->macDA_ = find_mac_addr(tmp_flow_queue->BWA_Next_Hop);
        hdr_lwx::access(tmp_pkt)->from_node_id()  = tmp_flow_queue->from_node_id;
        hdr_lwx::access(tmp_pkt)->bwa_flow_type() = tmp_flow_queue->BWA_Flow_Type;
        
   if (Scheduler::instance().clock() - hdr_lwx::access(tmp_pkt)->sendtime_ > hdr_lwx::access(tmp_pkt)->maxl_ )
      {     
        //drop packet added by pandeli
        drop(tmp_pkt, "DROP_RTR_TTL");
	}
	else {
        // send down packet
        downtarget_->recv(tmp_pkt,(Handler*) 0);
        }
        
      }
    }
  }
}
//**************** CODE by PANDELI ********************
void LWX::trans_rs_to_rs_nontp_dl(int node_id, vector <Flow_Queue*> &tmp_flow_list)
{
  for( int i=0; i <  tmp_flow_list.size(); i++)
  {     
    
    Flow_Queue *tmp_flow_queue = tmp_flow_list.at(i);
    
    if(tmp_flow_queue->from_node_id == node_id)
    {
      // log test msg

      while(tmp_flow_queue->queue_length()!=0)
      {
        // get packet
        Packet *tmp_pkt=tmp_flow_queue->pkt_front();

        // decrease BWA_Total
        if((tmp_flow_queue->BWA_Total_Byte - hdr_cmn::access(tmp_pkt)->size())<0 )
        {
          tmp_flow_queue->BWA_Backoff = tmp_flow_queue->BWA_Total_Byte; 
          break; 
        }
        tmp_flow_queue->BWA_Total_Byte -= hdr_cmn::access(tmp_pkt)->size();
     
        // pop this packet
        tmp_flow_queue->pkt_pop();
        
        // get this pkt's next hop's mac
        // chang pkt direction from up to down
        // and change this pkt's mac_col->macDA_ 
        // and change lwx from_node_id and bwa_flow_type
        hdr_cmn::access(tmp_pkt)->direction() = hdr_cmn::DOWN;
        hdr_mac::access(tmp_pkt)->macDA_ = find_mac_addr(tmp_flow_queue->BWA_Next_Hop);
        hdr_lwx::access(tmp_pkt)->from_node_id()  = tmp_flow_queue->from_node_id;
        hdr_lwx::access(tmp_pkt)->bwa_flow_type() = tmp_flow_queue->BWA_Flow_Type;
        
   if (Scheduler::instance().clock() - hdr_lwx::access(tmp_pkt)->sendtime_ > hdr_lwx::access(tmp_pkt)->maxl_ )
      {   
        //drop packet 
        drop(tmp_pkt, "DROP_RTR_TTL");
	}
	else {
        // send down packet
        downtarget_->recv(tmp_pkt,(Handler*) 0);
        }
        
      }
    }
  }
}

void LWX::trans_rs_to_rs_nontp_ul(int node_id, vector <Flow_Queue*> &tmp_flow_list)
{
  for( int i=0; i <  tmp_flow_list.size(); i++)
  {     
    
    Flow_Queue *tmp_flow_queue = tmp_flow_list.at(i);
    
    if(tmp_flow_queue->from_node_id == node_id)
    {
      // log test msg

      while(tmp_flow_queue->queue_length()!=0)
      {
        // get packet
        Packet *tmp_pkt=tmp_flow_queue->pkt_front();

        // decrease BWA_Total
        if((tmp_flow_queue->BWA_Total_Byte - hdr_cmn::access(tmp_pkt)->size())<0 )
        {
          tmp_flow_queue->BWA_Backoff = tmp_flow_queue->BWA_Total_Byte; 
          break; 
        }
        tmp_flow_queue->BWA_Total_Byte -= hdr_cmn::access(tmp_pkt)->size();
     
        // pop this packet
        tmp_flow_queue->pkt_pop();
        
        // get this pkt's next hop's mac
        // chang pkt direction from up to down
        // and change this pkt's mac_col->macDA_ 
        // and change lwx from_node_id and bwa_flow_type
        hdr_cmn::access(tmp_pkt)->direction() = hdr_cmn::DOWN;
        hdr_mac::access(tmp_pkt)->macDA_ = find_mac_addr(tmp_flow_queue->BWA_Next_Hop);
        hdr_lwx::access(tmp_pkt)->from_node_id()  = tmp_flow_queue->from_node_id;
        hdr_lwx::access(tmp_pkt)->bwa_flow_type() = tmp_flow_queue->BWA_Flow_Type;
        
   if (Scheduler::instance().clock() - hdr_lwx::access(tmp_pkt)->sendtime_ > hdr_lwx::access(tmp_pkt)->maxl_ )
      { 
        //drop packet 
        drop(tmp_pkt, "DROP_RTR_TTL");
	}
	else {
        // send down packet
        downtarget_->recv(tmp_pkt,(Handler*) 0);
        }
        
      }
    }
  }
}

//************* END code by PANDELI ************************************  


/////////////////////
//  Timer Handler
/////////////////////

//
// timer flow chart:
// 
//   sec 0 : LWX::LWX() : Mac() => frame_timer->start(0) => call Frame_LWXTimer::handle
//   sec 0 : Frame_LWXTimer::handle => call mac->handle_rtg(node_id)
//   sec 0 : void LWX::handle_rtg(int node_id) => exec bandwidth allocation
//   sec 0 : void LWX::handle_rtg(int node_id) call rtg_timer->start(RTG_PERIOD) 
//           to wait RTG_PERIOD time 0.1ms
//   sec 0.1ms : rtg timeout => call RTG_LWXTimer::handle(Event*)
//   sec 0.1ms : RTG_LWXTimer::handle(Event*) call mac->handle_flow_trans(node_id)
//   sec 0.1ms : void LWX::handle_flow_trans(int node_id) call 
//               trans_downlink(node_id); and trans_uplink(node_id); to trans data
//   sec 0.1ms : void LWX::handle_flow_trans(int node_id) call 
//               frame_timer->start(FRAME_PERIOD); to wait FRAME_PERIOD time 4.9ms
//   sec 5ms   : FRAME timeout =>  call Frame_LWXTimer::handle
//   
//                 (repeat)
// 

void LWX::handle_flow_trans(int node_id)
{
  
  int src_node_id   = Node::get_node_by_address(index_)->nodeid();

  // create pointer to downlink & uplink queue to allocation bandwidth efficiencially
  vector <Flow_Queue*> tmp_ss_to_bs;
  vector <Flow_Queue*> tmp_bs_to_ss;
  vector <Flow_Queue*> tmp_bs_to_rs;
  vector <Flow_Queue*> tmp_rs_to_bs;
  vector <Flow_Queue*> tmp_ss_to_rs_tp;
  vector <Flow_Queue*> tmp_ss_to_rs_nontp;
  vector <Flow_Queue*> tmp_rs_to_ss_tp;
  vector <Flow_Queue*> tmp_rs_to_ss_nontp;
  vector <Flow_Queue*> tmp_rs_to_rs_nontp_dl; // added by pandeli
  vector <Flow_Queue*> tmp_rs_to_rs_nontp_ul; // added by pandeli
  vector <Flow_Queue*> tmp_rs_to_rs_nontp_dl_odd; // added by pandeli
  vector <Flow_Queue*> tmp_rs_to_rs_nontp_ul_odd; // added by pandeli
  
  // classify flows of the node
  // whose node id is src_node_id into corresponding flow queue
  // and call the corresponding transmission function

  for(int i=0; i < lwx_node_list.size(); i++)
  {
    if(lwx_node_list.at(i).node_id == src_node_id)
    { 
      LWX_Node *tmp_node = &lwx_node_list.at(i);
      for( int j=0; j <  tmp_node->ctrl_flow.size(); j++)
      { 
        switch(tmp_node->ctrl_flow.at(j).BWA_Flow_Type)
        {
          case FLOW_TYPE_BS_TO_SS:
            tmp_bs_to_ss.push_back(&tmp_node->ctrl_flow.at(j));
            break;
          case FLOW_TYPE_SS_TO_BS:
            tmp_ss_to_bs.push_back(&tmp_node->ctrl_flow.at(j));
            break;
          case FLOW_TYPE_BS_TO_RS:
            tmp_bs_to_rs.push_back(&tmp_node->ctrl_flow.at(j));
            break;
          case FLOW_TYPE_RS_TO_BS:
            tmp_rs_to_bs.push_back(&tmp_node->ctrl_flow.at(j));
            break;
          case FLOW_TYPE_SS_TO_RS_TP:
            tmp_ss_to_rs_tp.push_back(&tmp_node->ctrl_flow.at(j));
            break;
          case FLOW_TYPE_SS_TO_RS_NONTP:
            tmp_ss_to_rs_nontp.push_back(&tmp_node->ctrl_flow.at(j));
            break;
          case FLOW_TYPE_RS_TO_SS_TP:
            tmp_rs_to_ss_tp.push_back(&tmp_node->ctrl_flow.at(j));
            break;
          case FLOW_TYPE_RS_TO_SS_NONTP:
            tmp_rs_to_ss_nontp.push_back(&tmp_node->ctrl_flow.at(j));
            break;
          case FLOW_TYPE_RS_TO_RS_NONTP_DL:    // added by Pandeli 
            tmp_rs_to_rs_nontp_dl.push_back(&tmp_node->ctrl_flow.at(j));  
            break; 	  
          case FLOW_TYPE_RS_TO_RS_NONTP_UL:    // added by Pandeli 
            tmp_rs_to_rs_nontp_ul.push_back(&tmp_node->ctrl_flow.at(j));  
            break;
          case FLOW_TYPE_RS_TO_RS_NONTP_DL_ODD:    // added by Pandeli 
            tmp_rs_to_rs_nontp_dl_odd.push_back(&tmp_node->ctrl_flow.at(j));  
            break; 	  
          case FLOW_TYPE_RS_TO_RS_NONTP_UL_ODD:    // added by Pandeli 
            tmp_rs_to_rs_nontp_ul_odd.push_back(&tmp_node->ctrl_flow.at(j));  
            break;
          default:
            lwx_err_msg("in handle_transmission", "wrong flow type");
            break;          
        }
      }
    }
  }
  
  // send packet
  trans_bs_to_ss(node_id, tmp_bs_to_ss);
  trans_ss_to_bs(node_id, tmp_ss_to_bs);
  trans_bs_to_rs(node_id, tmp_bs_to_rs);
  trans_rs_to_bs(node_id, tmp_rs_to_bs);
  trans_ss_to_rs_tp(node_id, tmp_ss_to_rs_tp);
  trans_ss_to_rs_nontp(node_id, tmp_ss_to_rs_nontp);
  trans_rs_to_ss_tp(node_id, tmp_rs_to_ss_tp);
  trans_rs_to_ss_nontp(node_id, tmp_rs_to_ss_nontp);  
  trans_rs_to_rs_nontp_dl(node_id, tmp_rs_to_rs_nontp_dl); // added by Pandeli
  trans_rs_to_rs_nontp_ul(node_id, tmp_rs_to_rs_nontp_ul); // added by Pandeli
  trans_rs_to_rs_nontp_dl(node_id, tmp_rs_to_rs_nontp_dl_odd); // added by Pandeli
  trans_rs_to_rs_nontp_ul(node_id, tmp_rs_to_rs_nontp_ul_odd); // added by Pandeli

  
  // enter here in next frame dealling time
  frame_timer->start(FRAME_PERIOD);
  
}


void LWX::handle_rtg(int node_id)
{
  // allocate bandwidth
  // only bs can execute bandwidth allocation 
  if(is_bs(node_id))
  {
    bandwidth_allocation(node_id);  
  }

  // finish bandwidth allocation
  // enact RTG Timer=> after RTG time, start transmission packet of each flow
  
  rtg_timer->start(RTG_PERIOD);
}

/////////////////////
//  creat and send lwx mac layer packet
/////////////////////

bool LWX::send_Ctrl_Pkt() {
  
  Packet *ctrl_pkt= Packet::alloc();	
 
  
  hdr_cmn *cmn_col  = hdr_cmn::access(ctrl_pkt);
  hdr_mac *mac_col  = hdr_mac::access(ctrl_pkt);
  hdr_lwx *lwx_col  = hdr_lwx::access(ctrl_pkt);
  hdr_ip  *ip_col   = hdr_ip::access(ctrl_pkt);

  // set common layer column
	cmn_col->txtime() = 0.0001;
	cmn_col->uid()    = 0;
  cmn_col->ptype()  = PT_LWX;
  cmn_col->size()   = 24; // iHHK
  cmn_col->iface()  = UNKN_IFACE.value();
  cmn_col->error()  = 0; 

  // set mac layer column
  mac_col->macDA_   = 2;
  mac_col->macSA_   = index_;	 // sender's mac address  

  // set lwx layer column
  
  lwx_col->src_ip() = 153;
  
  
  lwx_test_msg ("LWX::send_Ctrl_Pkt()");
  // eXʥ]

  downtarget_->recv(ctrl_pkt);
	
  return 1;
}

/////////////////////
//  Timers start:
/////////////////////

//    in this section, I implement the virtual funciton here
//    # start(time):   execute the LWXTimer::handle() after "time" sec
//    # restart(time): execute the LWXTimer::handle() after "time" sec
//                     if not busy
//    # stop(time):    cancel this event
                       
void LWXTimer::start(double time)
{
  Scheduler &s = Scheduler::instance();

  assert(busy_ == 0);
  busy_ = 1;
  paused_ = 0;
  stime = s.clock();
  wtime = time;
  assert(wtime >= 0.0);
  event.uid_ = 0;
  s.schedule(this, &event, wtime);
}
 
void LWXTimer::stop(void)
{ 
  Scheduler &s = Scheduler::instance();

  assert(busy_);
  if(paused_ == 0)
    s.cancel(&event);
}

void LWXTimer::handle(Event* e)
{
}

// Frame_LWXTimer

void Frame_LWXTimer::start(double time){	
  Scheduler &s = Scheduler::instance();	
  assert(busy_ == 0);	
  busy_ = 1;
  paused_ = 1;

  stime = s.clock();
  wtime = time;
  assert(wtime >= 0.0);
  event.uid_ = 0;
  s.schedule(this, &event, wtime);
}

void Frame_LWXTimer::stop(){
  Scheduler &s = Scheduler::instance();
  assert(busy_);
  if(paused_ == 0)
    s.cancel(&event);
}
// pplong:
//   While timeout, the Timer will trigger this handle
//   and this hanlde will call other function.
//   In this case, this handler will call LWX's handle_flow_trans().
//   By the way, the wiatHandler() is not formal name, you can define prevision,
//   and call any function here.
void Frame_LWXTimer::handle(Event*){ 
  busy_=0;
  paused_=0;
  stime=0.0;
  wtime=0.0;
  // Ishandle_rtgA}lbandwidth allocation
  mac->handle_rtg(Node::get_node_by_address(mac->addr())->nodeid());
}

// RTG_LWXTimer

void RTG_LWXTimer::start(double time){	
  Scheduler &s = Scheduler::instance();	
  assert(busy_ == 0);	
  busy_ = 1;
  paused_ = 1;

  stime = s.clock();
  wtime = time;
  assert(wtime >= 0.0);
  event.uid_ = 0;
  s.schedule(this, &event, wtime);
}

void RTG_LWXTimer::stop(){
  Scheduler &s = Scheduler::instance();
  assert(busy_);
  if(paused_ == 0)
    s.cancel(&event);
}
// pplong:
//   While timeout, the Timer will trigger this handle
//   and this hanlde will call other function.
//   In this case, this handler will call LWX's handle_flow_trans().
//   By the way, the wiatHandler() is not formal name, you can define prevision,
//   and call any function here.
void RTG_LWXTimer::handle(Event*){ 
  busy_=0;
  paused_=0;
  stime=0.0;
  wtime=0.0;
  // Ishandle_flow_transA}lwҤtWeǫʥ]
  mac->handle_flow_trans(Node::get_node_by_address(mac->addr())->nodeid());	
}

/////////////////////
// Timer end
/////////////////////

